#!/usr/bin/env python

import io
import os
import re
import sys
import time
import socket
import locale
import logging
import argparse
from html import unescape
from http import cookiejar
from importlib import import_module
from urllib import (
    request,
    parse,
    error
)
from multiprocessing.dummy import Pool

from lulu import config
from lulu.util import log, term
from lulu.version import __version__
from lulu import json_output as json_output_
from lulu.util.strings import get_filename

sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf8')


dry_run = False
json_output = False
force = False
player = None
extractor_proxy = None
cookies = None
output_filename = None


if sys.stdout.isatty():
    default_encoding = sys.stdout.encoding.lower()
else:
    default_encoding = locale.getpreferredencoding().lower()


def rc4(key, data):
    # all encryption algo should work on bytes
    assert type(key) == type(data) and type(key) == type(b'')
    state = list(range(256))
    j = 0
    for i in range(256):
        j += state[i] + key[i % len(key)]
        j &= 0xff
        state[i], state[j] = state[j], state[i]

    i = 0
    j = 0
    out_list = []
    for char in data:
        i += 1
        i &= 0xff
        j += state[i]
        j &= 0xff
        state[i], state[j] = state[j], state[i]
        prn = state[(state[i] + state[j]) & 0xff]
        out_list.append(char ^ prn)

    return bytes(out_list)


def general_m3u8_extractor(url, headers=config.FAKE_HEADERS):
    m3u8_list = get_content(url, headers=headers).split('\n')
    urls = []
    for line in m3u8_list:
        line = line.strip()
        if line and not line.startswith('#'):
            if line.startswith('http'):
                urls.append(line)
            else:
                seg_url = parse.urljoin(url, line)
                urls.append(seg_url)
    return urls


def maybe_print(*s):
    try:
        print(*s)
    except Exception:
        pass


def tr(s):
    if default_encoding == 'utf-8':
        return s
    else:
        return s
        # return str(s.encode('utf-8'))[2:-1]


def match1(text, *patterns):
    """Scans through a string for substrings matched some patterns
    (first-subgroups only).

    Args:
        text: A string to be scanned.
        patterns: Arbitrary number of regex patterns.

    Returns:
        When only one pattern is given, returns a string
        (None if no match found).
        When more than one pattern are given, returns a list of strings
        ([] if no match found).
    """

    if len(patterns) == 1:
        pattern = patterns[0]
        match = re.search(pattern, text)
        if match:
            return match.group(1)
        else:
            return None
    else:
        ret = []
        for pattern in patterns:
            match = re.search(pattern, text)
            if match:
                ret.append(match.group(1))
        return ret


def matchall(text, patterns):
    """Scans through a string for substrings matched some patterns.

    Args:
        text: A string to be scanned.
        patterns: a list of regex pattern.

    Returns:
        a list if matched. empty if not.
    """

    ret = []
    for pattern in patterns:
        match = re.findall(pattern, text)
        ret += match

    return ret


def launch_player(player, urls, refer=''):
    import subprocess
    import shlex
    player = shlex.split(player)
    params = []
    ua = config.FAKE_HEADERS['User-Agent']
    if player[0] == 'mpv':
        params.append('--no-ytdl')
        params.extend(['--user-agent', ua])
        if refer:
            params.extend([
                '--http-header-fields',
                'referer: {}'.format(refer)
            ])
    subprocess.call(player + params + list(urls))


def parse_query_param(url, param):
    """Parses the query string of a URL and returns the value of a parameter.

    Args:
        url: A URL.
        param: A string representing the name of the parameter.

    Returns:
        The value of the parameter.
    """

    try:
        return parse.parse_qs(parse.urlparse(url).query)[param][0]
    except Exception:
        return None


def unicodize(text):
    return re.sub(
        r'\\u([0-9A-Fa-f][0-9A-Fa-f][0-9A-Fa-f][0-9A-Fa-f])',
        lambda x: chr(int(x.group(0)[2:], 16)),
        text
    )


def ungzip(data):
    """Decompresses data for Content-Encoding: gzip.
    """
    from io import BytesIO
    import gzip
    buffer = BytesIO(data)
    f = gzip.GzipFile(fileobj=buffer)
    return f.read()


def undeflate(data):
    """Decompresses data for Content-Encoding: deflate.
    (the zlib compression is used.)
    """
    import zlib
    decompressobj = zlib.decompressobj(-zlib.MAX_WBITS)
    return decompressobj.decompress(data)+decompressobj.flush()


def get_location(url):
    logging.debug('get_location: %s' % url)

    response = request.urlopen(url)
    # urllib will follow redirections and it's too much code to tell urllib
    # not to do that
    return response.geturl()


def urlopen_with_retry(*args, **kwargs):
    import ssl
    context = ssl._create_unverified_context()
    retry_time = 3
    for i in range(retry_time):
        try:
            return request.urlopen(*args, context=context, **kwargs)
        except socket.timeout as e:
            logging.debug('request attempt %s timeout' % str(i + 1))
            if i + 1 == retry_time:
                raise e
        # try to tackle youku CDN fails
        except error.HTTPError as http_error:
            logging.debug('HTTP Error with code{}'.format(http_error.code))
            if i + 1 == retry_time:
                raise http_error


def get_content(url, headers=config.FAKE_HEADERS, decoded=True):
    """Gets the content of a URL via sending a HTTP GET request.

    Args:
        url: A URL.
        headers: Request headers used by the client.
        decoded: Whether decode the response body using UTF-8 or the charset
        specified in Content-Type.

    Returns:
        The content as a string.
    """

    logging.debug('get_content: %s' % url)

    req = request.Request(url, headers=headers)
    if cookies:
        cookies.add_cookie_header(req)
        req.headers.update(req.unredirected_hdrs)

    response = urlopen_with_retry(req)
    data = response.read()

    # Handle HTTP compression for gzip and deflate (zlib)
    content_encoding = response.getheader('Content-Encoding')
    if content_encoding == 'gzip':
        data = ungzip(data)
    elif content_encoding == 'deflate':
        data = undeflate(data)

    # Decode the response body
    if decoded:
        charset = match1(
            response.getheader('Content-Type'), r'charset=([\w-]+)'
        )
        if charset is not None:
            data = data.decode(charset)
        else:
            data = data.decode('utf-8', 'ignore')

    return data


def post_content(url, headers={}, post_data={}, decoded=True):
    """Post the content of a URL via sending a HTTP POST request.

    Args:
        url: A URL.
        headers: Request headers used by the client.
        decoded: Whether decode the response body using UTF-8 or the charset
        specified in Content-Type.

    Returns:
        The content as a string.
    """

    logging.debug('post_content: %s \n post_data: %s' % (url, post_data))

    req = request.Request(url, headers=headers)
    if cookies:
        cookies.add_cookie_header(req)
        req.headers.update(req.unredirected_hdrs)
    post_data_enc = bytes(parse.urlencode(post_data), 'utf-8')
    response = urlopen_with_retry(req, data=post_data_enc)
    data = response.read()

    # Handle HTTP compression for gzip and deflate (zlib)
    content_encoding = response.getheader('Content-Encoding')
    if content_encoding == 'gzip':
        data = ungzip(data)
    elif content_encoding == 'deflate':
        data = undeflate(data)

    # Decode the response body
    if decoded:
        charset = match1(
            response.getheader('Content-Type'), r'charset=([\w-]+)'
        )
        if charset is not None:
            data = data.decode(charset)
        else:
            data = data.decode('utf-8')

    return data


def url_size(url, faker=False, headers={}):
    if faker:
        response = urlopen_with_retry(
            request.Request(url, headers=config.FAKE_HEADERS)
        )
    elif headers:
        response = urlopen_with_retry(request.Request(url, headers=headers))
    else:
        response = urlopen_with_retry(url)

    size = response.headers['content-length']
    return int(size) if size is not None else float('inf')


def urls_size(urls, faker=False, headers={}):
    return sum([url_size(url, faker=faker, headers=headers) for url in urls])


def get_head(url, headers={}, get_method='HEAD'):
    logging.debug('get_head: %s' % url)

    if headers:
        req = request.Request(url, headers=headers)
    else:
        req = request.Request(url)
    req.get_method = lambda: get_method
    res = urlopen_with_retry(req)
    return dict(res.headers)


def url_info(url, faker=False, headers={}):
    logging.debug('url_info: %s' % url)

    if faker:
        response = urlopen_with_retry(
            request.Request(url, headers=config.FAKE_HEADERS)
        )
    elif headers:
        response = urlopen_with_retry(request.Request(url, headers=headers))
    else:
        response = urlopen_with_retry(request.Request(url))

    headers = response.headers

    type = headers['content-type']
    if type == 'image/jpg; charset=UTF-8' or type == 'image/jpg':
        type = 'audio/mpeg'  # fix for netease
    mapping = {
        'video/3gpp': '3gp',
        'video/f4v': 'flv',
        'video/mp4': 'mp4',
        'video/MP2T': 'ts',
        'video/quicktime': 'mov',
        'video/webm': 'webm',
        'video/x-flv': 'flv',
        'video/x-ms-asf': 'asf',
        'audio/mp4': 'mp4',
        'audio/mpeg': 'mp3',
        'audio/wav': 'wav',
        'audio/x-wav': 'wav',
        'audio/wave': 'wav',
        'image/jpeg': 'jpg',
        'image/png': 'png',
        'image/gif': 'gif',
        'application/pdf': 'pdf',
    }
    if type in mapping:
        ext = mapping[type]
    elif '.' in url:
        type = ext = url.split('.')[-1]
    else:
        type = None
        if headers['content-disposition']:
            try:
                filename = parse.unquote(
                    match1(
                        headers['content-disposition'], r'filename="?([^"]+)"?'
                    )
                )
                if len(filename.split('.')) > 1:
                    ext = filename.split('.')[-1]
                else:
                    ext = None
            except Exception:
                ext = None
        else:
            ext = None

    if headers['transfer-encoding'] != 'chunked':
        size = headers['content-length'] and int(headers['content-length'])
    else:
        size = None

    return type, ext, size


def url_locations(urls, faker=False, headers={}):
    locations = []
    for url in urls:
        logging.debug('url_locations: %s' % url)

        if faker:
            response = urlopen_with_retry(
                request.Request(url, headers=config.FAKE_HEADERS)
            )
        elif headers:
            response = urlopen_with_retry(
                request.Request(url, headers=headers)
            )
        else:
            response = urlopen_with_retry(request.Request(url))

        locations.append(response.url)
    return locations


def url_save(
    url, filepath, bar, refer=None, is_part=False, faker=False,
    headers=None, timeout=None, **kwargs
):
    tmp_headers = headers.copy() if headers is not None else {}
    # When a referer specified with param refer,
    # the key must be 'Referer' for the hack here
    if refer is not None:
        tmp_headers['Referer'] = refer
    file_size = url_size(url, faker=faker, headers=tmp_headers)

    if os.path.exists(filepath):
        if not force and file_size == os.path.getsize(filepath):
            if not is_part:
                if bar:
                    bar.done()
                print(
                    'Skipping {}: file already exists'.format(
                        tr(os.path.basename(filepath))
                    )
                )
            else:
                if bar:
                    bar.update_received(file_size)
            return
        else:
            if not is_part:
                if bar:
                    bar.done()
                print('Overwriting %s' % tr(os.path.basename(filepath)), '...')
    elif not os.path.exists(os.path.dirname(filepath)):
        os.mkdir(os.path.dirname(filepath))

    temp_filepath = filepath + '.download' if file_size != float('inf') \
        else filepath
    received = 0
    if not force:
        open_mode = 'ab'

        if os.path.exists(temp_filepath):
            received += os.path.getsize(temp_filepath)
            if bar:
                bar.update_received(os.path.getsize(temp_filepath))
    else:
        open_mode = 'wb'

    if received < file_size:
        if faker:
            tmp_headers = config.FAKE_HEADERS
        '''
        if parameter headers passed in, we have it copied as tmp_header
        elif headers:
            headers = headers
        else:
            headers = {}
        '''
        if received:
            tmp_headers['Range'] = 'bytes=' + str(received) + '-'
        if refer:
            tmp_headers['Referer'] = refer

        if timeout:
            response = urlopen_with_retry(
                request.Request(url, headers=tmp_headers), timeout=timeout
            )
        else:
            response = urlopen_with_retry(
                request.Request(url, headers=tmp_headers)
            )
        try:
            range_start = int(
                response.headers[
                    'content-range'
                ][6:].split('/')[0].split('-')[0]
            )
            end_length = int(
                response.headers['content-range'][6:].split('/')[1]
            )
            range_length = end_length - range_start
        except Exception:
            content_length = response.headers['content-length']
            range_length = int(content_length) if content_length is not None \
                else float('inf')

        if file_size != received + range_length:
            received = 0
            if bar:
                bar.received = 0
            open_mode = 'wb'

        with open(temp_filepath, open_mode) as output:
            while True:
                buffer = None
                try:
                    buffer = response.read(1024 * 256)
                except socket.timeout:
                    pass
                if not buffer:
                    if received == file_size:  # Download finished
                        break
                    # Unexpected termination. Retry request
                    tmp_headers['Range'] = 'bytes=' + str(received) + '-'
                    response = urlopen_with_retry(
                        request.Request(url, headers=tmp_headers)
                    )
                    continue
                output.write(buffer)
                received += len(buffer)
                if bar:
                    bar.update_received(len(buffer))

    assert received == os.path.getsize(temp_filepath), '{} == {} == {}'.format(
        received, os.path.getsize(temp_filepath), temp_filepath
    )

    if os.access(filepath, os.W_OK):
        # on Windows rename could fail if destination filepath exists
        os.remove(filepath)
    os.rename(temp_filepath, filepath)


class SimpleProgressBar:
    term_size = term.get_terminal_size()[1]

    def __init__(self, total_size, total_pieces=1):
        self.displayed = False
        self.total_size = total_size
        self.total_pieces = total_pieces
        self.current_piece = 1
        self.received = 0
        self.speed = ''
        self.last_updated = time.time()

        total_pieces_len = len(str(total_pieces))
        # 38 is the size of all statically known size in self.bar
        total_str = '%5s' % round(self.total_size / 1048576, 1)
        total_str_width = max(len(total_str), 5)
        self.bar_size = self.term_size - 28 - 2 * total_pieces_len \
            - 2 * total_str_width
        self.bar = '{:>4}%% ({:>%s}/%sMB) ├{:─<%s}┤[{:>%s}/{:>%s}] {}' % (
            total_str_width, total_str, self.bar_size, total_pieces_len,
            total_pieces_len
        )

    def update(self):
        self.displayed = True
        bar_size = self.bar_size
        percent = round(self.received * 100 / self.total_size, 1)
        if percent >= 100:
            percent = 100
        dots = bar_size * int(percent) // 100
        plus = int(percent) - dots // bar_size * 100
        if plus > 0.8:
            plus = '█'
        elif plus > 0.4:
            plus = '>'
        else:
            plus = ''
        bar = '█' * dots + plus
        bar = self.bar.format(
            percent, round(self.received / 1048576, 1), bar,
            self.current_piece, self.total_pieces, self.speed
        )
        sys.stdout.write('\r' + bar)
        sys.stdout.flush()

    def update_received(self, n):
        self.received += n
        time_diff = time.time() - self.last_updated
        bytes_ps = n / time_diff if time_diff else 0
        if bytes_ps >= 1024 ** 3:
            self.speed = '{:4.0f} GB/s'.format(bytes_ps / 1024 ** 3)
        elif bytes_ps >= 1024 ** 2:
            self.speed = '{:4.0f} MB/s'.format(bytes_ps / 1024 ** 2)
        elif bytes_ps >= 1024:
            self.speed = '{:4.0f} kB/s'.format(bytes_ps / 1024)
        else:
            self.speed = '{:4.0f}  B/s'.format(bytes_ps)
        self.last_updated = time.time()
        self.update()

    def update_piece(self, n):
        self.current_piece = n

    def done(self):
        if self.displayed:
            print()
            self.displayed = False


class PiecesProgressBar:
    def __init__(self, total_size, total_pieces=1):
        self.displayed = False
        self.total_size = total_size
        self.total_pieces = total_pieces
        self.current_piece = 1
        self.received = 0

    def update(self):
        self.displayed = True
        bar = '{0:>5}%[{1:<40}] {2}/{3}'.format(
            '', '=' * 40, self.current_piece, self.total_pieces
        )
        sys.stdout.write('\r' + bar)
        sys.stdout.flush()

    def update_received(self, n):
        self.received += n
        self.update()

    def update_piece(self, n):
        self.current_piece = n

    def done(self):
        if self.displayed:
            print()
            self.displayed = False


class DummyProgressBar:
    def __init__(self, *args):
        pass

    def update_received(self, n):
        pass

    def update_piece(self, n):
        pass

    def done(self):
        pass


def get_output_filename(urls, title, ext, output_dir, merge):
    # lame hack for the --output-filename option
    global output_filename
    if output_filename:
        if ext:
            return output_filename + '.' + ext
        return output_filename

    merged_ext = ext
    if (len(urls) > 1) and merge:
        from .processor.ffmpeg import has_ffmpeg_installed
        if ext in ['flv', 'f4v']:
            if has_ffmpeg_installed():
                merged_ext = 'mp4'
            else:
                merged_ext = 'flv'
        elif ext == 'mp4':
            merged_ext = 'mp4'
        elif ext == 'ts':
            if has_ffmpeg_installed():
                merged_ext = 'mkv'
            else:
                merged_ext = 'ts'
    return '{}.{}'.format(title, merged_ext)


def download_urls(
    urls, title, ext, total_size, output_dir='.', refer=None, merge=True,
    faker=False, headers={}, thread=0, **kwargs
):
    assert urls
    if json_output:
        json_output_.download_urls(
            urls=urls, title=title, ext=ext, total_size=total_size,
            refer=refer
        )
        return
    if dry_run:
        print('Real URLs:\n{}'.format('\n'.join(urls)))
        return

    if player:
        launch_player(player, urls, refer=refer)
        return

    if not total_size:
        try:
            total_size = urls_size(urls, faker=faker, headers=headers)
        except Exception:
            import traceback
            traceback.print_exc(file=sys.stdout)
            pass

    title = tr(get_filename(title))
    output_filename = get_output_filename(urls, title, ext, output_dir, merge)
    output_filepath = os.path.join(output_dir, output_filename)

    if total_size:
        if not force and os.path.exists(output_filepath) \
                and os.path.getsize(output_filepath) >= total_size * 0.9:
            print('Skipping {}: file already exists'.format(output_filepath))
            print()
            return
        bar = SimpleProgressBar(total_size, len(urls))
    else:
        bar = PiecesProgressBar(total_size, len(urls))

    if len(urls) == 1:
        url = urls[0]
        print('Downloading {} ...'.format(tr(output_filename)))
        bar.update()
        url_save(
            url, output_filepath, bar, refer=refer, faker=faker,
            headers=headers, **kwargs
        )
        bar.done()
    else:
        print('Downloading {}.{} ...'.format(tr(title), ext))
        parts = [''] * len(urls)
        bar.update()
        piece = 0

        def _download(url):
            # Closure
            nonlocal piece
            index = urls.index(url)
            filename = '{}[{:0>2d}].{}'.format(title, index, ext)
            filepath = os.path.join(output_dir, filename)
            parts[index] = filepath  # 防止多线程环境下文件顺序会乱
            piece += 1
            bar.update_piece(piece)
            url_save(
                url, filepath, bar, refer=refer, is_part=True, faker=faker,
                headers=headers, **kwargs
            )
        if thread:
            with Pool(processes=thread) as pool:
                pool.map(_download, urls)
        else:
            for url in urls:
                _download(url)
        bar.done()

        if not merge:
            print()
            return

        if 'av' in kwargs and kwargs['av']:
            from .processor.ffmpeg import has_ffmpeg_installed
            if has_ffmpeg_installed():
                from .processor.ffmpeg import ffmpeg_concat_av
                ret = ffmpeg_concat_av(parts, output_filepath, ext)
                print('Merged into {}'.format(output_filename))
                if ret == 0:
                    for part in parts:
                        os.remove(part)

        elif ext in ['flv', 'f4v']:
            try:
                from .processor.ffmpeg import has_ffmpeg_installed
                if has_ffmpeg_installed():
                    from .processor.ffmpeg import ffmpeg_concat_flv_to_mp4
                    ffmpeg_concat_flv_to_mp4(parts, output_filepath)
                else:
                    from .processor.join_flv import concat_flv
                    concat_flv(parts, output_filepath)
                print('Merged into {}'.format(output_filename))
            except Exception:
                raise
            else:
                for part in parts:
                    os.remove(part)

        elif ext == 'mp4':
            try:
                from .processor.ffmpeg import has_ffmpeg_installed
                if has_ffmpeg_installed():
                    from .processor.ffmpeg import ffmpeg_concat_mp4_to_mp4
                    ffmpeg_concat_mp4_to_mp4(parts, output_filepath)
                else:
                    from .processor.join_mp4 import concat_mp4
                    concat_mp4(parts, output_filepath)
                print('Merged into {}'.format(output_filename))
            except Exception:
                raise
            else:
                for part in parts:
                    os.remove(part)

        elif ext == 'ts':
            try:
                from .processor.ffmpeg import has_ffmpeg_installed
                if has_ffmpeg_installed():
                    from .processor.ffmpeg import ffmpeg_concat_ts_to_mkv
                    ffmpeg_concat_ts_to_mkv(parts, output_filepath)
                else:
                    from .processor.join_ts import concat_ts
                    concat_ts(parts, output_filepath)
                print('Merged into {}'.format(output_filename))
            except Exception:
                raise
            else:
                for part in parts:
                    os.remove(part)

        else:
            print("Can't merge {} files".format(ext))

    print()


def download_rtmp_url(
    url, title, ext, params={}, total_size=0, output_dir='.', refer=None,
    merge=True, faker=False
):
    assert url
    if dry_run:
        print('Real URL:\n%s\n' % [url])
        if params.get('-y', False):  # None or unset -> False
            print('Real Playpath:\n%s\n' % [params.get('-y')])
        return

    if player:
        from .processor.rtmpdump import play_rtmpdump_stream
        play_rtmpdump_stream(player, url, params)
        return

    from .processor.rtmpdump import (
        has_rtmpdump_installed, download_rtmpdump_stream
    )
    assert has_rtmpdump_installed(), 'RTMPDump not installed.'
    download_rtmpdump_stream(url,  title, ext, params, output_dir)


def download_url_ffmpeg(
    url, title, ext, params={}, total_size=0, output_dir='.', refer=None,
    merge=True, faker=False, stream=True, **kwargs
):
    assert url
    if dry_run:
        print('Real URL:\n%s\n' % [url])
        if params.get('-y', False):  # None or unset ->False
            print('Real Playpath:\n%s\n' % [params.get('-y')])
        return

    if player:
        launch_player(player, [url], refer=refer)
        return

    from .processor.ffmpeg import has_ffmpeg_installed, ffmpeg_download_stream
    assert has_ffmpeg_installed(), 'FFmpeg not installed.'

    global output_filename
    if output_filename:
        dotPos = output_filename.rfind('.')
        if dotPos > 0:
            title = output_filename[:dotPos]
            ext = output_filename[dotPos+1:]
        else:
            title = output_filename

    title = tr(get_filename(title))

    ffmpeg_download_stream(
        url, title, ext, params, output_dir, stream=stream, **kwargs
    )


def playlist_not_supported(name):
    def f(*args, **kwargs):
        raise NotImplementedError('Playlist is not supported for ' + name)
    return f


def print_info(site_info, title, type, size, **kwargs):
    if json_output:
        json_output_.print_info(
            site_info=site_info, title=title, type=type, size=size
        )
        return
    if type:
        type = type.lower()
    if type in ['3gp']:
        type = 'video/3gpp'
    elif type in ['asf', 'wmv']:
        type = 'video/x-ms-asf'
    elif type in ['flv', 'f4v']:
        type = 'video/x-flv'
    elif type in ['mkv']:
        type = 'video/x-matroska'
    elif type in ['mp3']:
        type = 'audio/mpeg'
    elif type in ['mp4']:
        type = 'video/mp4'
    elif type in ['mov']:
        type = 'video/quicktime'
    elif type in ['ts']:
        type = 'video/MP2T'
    elif type in ['webm']:
        type = 'video/webm'

    elif type in ['jpg']:
        type = 'image/jpeg'
    elif type in ['png']:
        type = 'image/png'
    elif type in ['gif']:
        type = 'image/gif'

    if type in ['video/3gpp']:
        type_info = '3GPP multimedia file (%s)' % type
    elif type in ['video/x-flv', 'video/f4v']:
        type_info = 'Flash video (%s)' % type
    elif type in ['video/mp4', 'video/x-m4v']:
        type_info = 'MPEG-4 video (%s)' % type
    elif type in ['video/MP2T']:
        type_info = 'MPEG-2 transport stream (%s)' % type
    elif type in ['video/webm']:
        type_info = 'WebM video (%s)' % type
    # elif type in ['video/ogg']:
    #    type_info = 'Ogg video (%s)' % type
    elif type in ['video/quicktime']:
        type_info = 'QuickTime video (%s)' % type
    elif type in ['video/x-matroska']:
        type_info = 'Matroska video (%s)' % type
    # elif type in ['video/x-ms-wmv']:
    #    type_info = 'Windows Media video (%s)' % type
    elif type in ['video/x-ms-asf']:
        type_info = 'Advanced Systems Format (%s)' % type
    # elif type in ['video/mpeg']:
    #    type_info = 'MPEG video (%s)' % type
    elif type in ['audio/mp4', 'audio/m4a']:
        type_info = 'MPEG-4 audio (%s)' % type
    elif type in ['audio/mpeg']:
        type_info = 'MP3 (%s)' % type
    elif type in ['audio/wav', 'audio/wave', 'audio/x-wav']:
        type_info = 'Waveform Audio File Format ({})'.format(type)

    elif type in ['image/jpeg']:
        type_info = 'JPEG Image (%s)' % type
    elif type in ['image/png']:
        type_info = 'Portable Network Graphics (%s)' % type
    elif type in ['image/gif']:
        type_info = 'Graphics Interchange Format (%s)' % type
    elif type in ['m3u8']:
        if 'm3u8_type' in kwargs:
            if kwargs['m3u8_type'] == 'master':
                type_info = 'M3U8 Master {}'.format(type)
        else:
            type_info = 'M3U8 Playlist {}'.format(type)
    else:
        type_info = 'Unknown type (%s)' % type

    maybe_print('Site:      ', site_info)
    maybe_print('Title:     ', unescape(tr(title)))
    print('Type:      ', type_info)
    if type != 'm3u8':
        print(
            'Size:      ', round(size / 1048576, 2),
            'MiB (' + str(size) + ' Bytes)'
        )
    if type == 'm3u8' and 'm3u8_url' in kwargs:
        print('M3U8 Url:   {}'.format(kwargs['m3u8_url']))
    print()


def mime_to_container(mime):
    mapping = {
        'video/3gpp': '3gp',
        'video/mp4': 'mp4',
        'video/webm': 'webm',
        'video/x-flv': 'flv',
    }
    if mime in mapping:
        return mapping[mime]
    else:
        return mime.split('/')[1]


def parse_host(host):
    """Parses host name and port number from a string.
    """
    if re.match(r'^(\d+)$', host) is not None:
        return ('0.0.0.0', int(host))
    if re.match(r'^(\w+)://', host) is None:
        host = '//' + host
    o = parse.urlparse(host)
    hostname = o.hostname or '0.0.0.0'
    port = o.port or 0
    return (hostname, port)


def set_proxy(proxy):
    proxy_handler = request.ProxyHandler({
        'http': '%s:%s' % proxy,
        'https': '%s:%s' % proxy,
    })
    opener = request.build_opener(proxy_handler)
    request.install_opener(opener)


def unset_proxy():
    proxy_handler = request.ProxyHandler({})
    opener = request.build_opener(proxy_handler)
    request.install_opener(opener)


def download_main(download, download_playlist, urls, playlist, **kwargs):
    for url in urls:
        if re.match(r'https?://', url) is None:
            url = 'http://' + url

        if playlist:
            download_playlist(url, **kwargs)
        else:
            download(url, **kwargs)


def load_cookies(cookiefile):
    global cookies
    try:
        cookies = cookiejar.MozillaCookieJar(cookiefile)
        cookies.load()
    except Exception:
        import sqlite3
        cookies = cookiejar.MozillaCookieJar()
        con = sqlite3.connect(cookiefile)
        cur = con.cursor()
        try:
            cur.execute("""SELECT host, path, isSecure, expiry, name, value
                        FROM moz_cookies""")
            for item in cur.fetchall():
                c = cookiejar.Cookie(
                    0, item[4], item[5], None, False, item[0],
                    item[0].startswith('.'), item[0].startswith('.'),
                    item[1], False, item[2], item[3], item[3] == '', None,
                    None, {},
                )
                cookies.set_cookie(c)
        except Exception:
            pass
        # TODO: Chromium Cookies
        # SELECT host_key, path, secure, expires_utc, name, encrypted_value
        # FROM cookies
        # http://n8henrie.com/2013/11/use-chromes-cookies-for-easier-downloading-with-python-requests/


def set_socks_proxy(proxy):
    try:
        import socks
        socks_proxy_addrs = proxy.split(':')
        socks.set_default_proxy(
            socks.SOCKS5,
            socks_proxy_addrs[0],
            int(socks_proxy_addrs[1])
        )
        socket.socket = socks.socksocket

        def getaddrinfo(*args):
            return [
                (socket.AF_INET, socket.SOCK_STREAM, 6, '', (args[0], args[1]))
            ]
        socket.getaddrinfo = getaddrinfo
    except ImportError:
        log.w(
            'Error importing PySocks library, socks proxy ignored.'
            'In order to use use socks proxy, please install PySocks.'
        )


def script_main(download, download_playlist, **kwargs):
    logging.basicConfig(format='[%(levelname)s] %(message)s')

    def print_version():
        version = __version__
        log.i(
            'version {}, a tiny downloader that scrapes the web.'.format(
                version
            )
        )

    parser = argparse.ArgumentParser(
        prog='lulu',
        usage='lulu [OPTION]... URL...',
        description='A tiny downloader that scrapes the web',
        add_help=False,
    )
    parser.add_argument(
        '-V', '--version', action='store_true',
        help='Print version and exit'
    )
    parser.add_argument(
        '-h', '--help', action='store_true',
        help='Print this help message and exit'
    )

    dry_run_grp = parser.add_argument_group(
        'Dry-run options', '(no actual downloading)'
    )
    dry_run_grp = dry_run_grp.add_mutually_exclusive_group()
    dry_run_grp.add_argument(
        '-i', '--info', action='store_true', help='Print extracted information'
    )
    dry_run_grp.add_argument(
        '-u', '--url', action='store_true',
        help='Print extracted information with URLs'
    )
    dry_run_grp.add_argument(
        '--json', action='store_true',
        help='Print extracted URLs in JSON format'
    )

    download_grp = parser.add_argument_group('Download options')
    download_grp.add_argument(
        '-n', '--no-merge', action='store_true', default=False,
        help='Do not merge video parts'
    )
    download_grp.add_argument(
        '--no-caption', action='store_true',
        help='Do not download captions (subtitles, lyrics, danmaku, ...)'
    )
    download_grp.add_argument(
        '-f', '--force', action='store_true', default=False,
        help='Force overwriting existing files'
    )
    download_grp.add_argument(
        '-F', '--format', metavar='STREAM_ID',
        help='Set video format to STREAM_ID'
    )
    download_grp.add_argument(
        '-O', '--output-filename', metavar='FILE', help='Set output filename'
    )
    download_grp.add_argument(
        '-o', '--output-dir', metavar='DIR', default='.',
        help='Set output directory'
    )
    download_grp.add_argument(
        '-p', '--player', metavar='PLAYER',
        help='Stream extracted URL to a PLAYER'
    )
    download_grp.add_argument(
        '-c', '--cookies', metavar='COOKIES_FILE',
        help='Load cookies.txt or cookies.sqlite'
    )
    download_grp.add_argument(
        '-t', '--timeout', metavar='SECONDS', type=int, default=600,
        help='Set socket timeout'
    )
    download_grp.add_argument(
        '-d', '--debug', action='store_true',
        help='Show traceback and other debug info'
    )
    download_grp.add_argument(
        '-I', '--input-file', metavar='FILE', type=argparse.FileType('r'),
        help='Read non-playlist URLs from FILE'
    )
    download_grp.add_argument(
        '-P', '--password', help='Set video visit password to PASSWORD'
    )
    download_grp.add_argument(
        '-l', '--playlist', action='store_true',
        help='Prefer to download a playlist'
    )
    download_grp.add_argument(
        '-T', '--thread', type=int, default=0,
        help=(
            'Use multithreading to download (only works for multiple-parts '
            'video)'
        )
    )

    proxy_grp = parser.add_argument_group('Proxy options')
    proxy_grp = proxy_grp.add_mutually_exclusive_group()
    proxy_grp.add_argument(
        '-x', '--http-proxy', metavar='HOST:PORT',
        help='Use an HTTP proxy for downloading'
    )
    proxy_grp.add_argument(
        '-y', '--extractor-proxy', metavar='HOST:PORT',
        help='Use an HTTP proxy for extracting only'
    )
    proxy_grp.add_argument(
        '--no-proxy', action='store_true', help='Never use a proxy'
    )
    proxy_grp.add_argument(
        '-s', '--socks-proxy', metavar='HOST:PORT',
        help='Use an SOCKS5 proxy for downloading'
    )

    download_grp.add_argument('--stream', help=argparse.SUPPRESS)
    download_grp.add_argument('--itag', help=argparse.SUPPRESS)

    parser.add_argument('URL', nargs='*', help=argparse.SUPPRESS)

    args = parser.parse_args()

    if args.help:
        print_version()
        parser.print_help()
        sys.exit()
    if args.version:
        print_version()
        sys.exit()

    if args.debug:
        # Set level of root logger to DEBUG
        logging.getLogger().setLevel(logging.DEBUG)

    global force
    global dry_run
    global json_output
    global player
    global extractor_proxy
    global output_filename

    output_filename = args.output_filename
    extractor_proxy = args.extractor_proxy

    info_only = args.info
    if args.url:
        dry_run = True
    if args.json:
        json_output = True
        # to fix extractors not use VideoExtractor
        dry_run = True
        info_only = False

    if args.cookies:
        load_cookies(args.cookies)

    caption = True
    stream_id = args.format or args.stream or args.itag
    if args.no_caption:
        caption = False
    if args.player:
        player = args.player
        caption = False

    if args.no_proxy:
        unset_proxy()
    else:
        if args.http_proxy:
            set_proxy(parse_host(args.http_proxy))
    if args.socks_proxy:
        set_socks_proxy(args.socks_proxy)

    URLs = []
    if args.input_file:
        logging.debug('you are trying to load urls from %s', args.input_file)
        if args.playlist:
            log.e(
                "reading playlist from a file is unsupported "
                "and won't make your life easier"
            )
            sys.exit(2)
        URLs.extend(args.input_file.read().splitlines())
        args.input_file.close()
    URLs.extend(args.URL)

    if not URLs:
        parser.print_help()
        sys.exit()

    socket.setdefaulttimeout(args.timeout)

    try:
        extra = {}
        if extractor_proxy:
            extra['extractor_proxy'] = extractor_proxy
        if stream_id:
            extra['stream_id'] = stream_id
        download_main(
            download, download_playlist,
            URLs, args.playlist,
            output_dir=args.output_dir, merge=not args.no_merge,
            info_only=info_only, json_output=json_output, caption=caption,
            password=args.password, thread=args.thread,
            **extra
        )
    except KeyboardInterrupt:
        if args.debug:
            raise
        else:
            sys.exit(1)
    except UnicodeEncodeError:
        if args.debug:
            raise
        log.e(
            '[error] oops, the current environment does not seem to support '
            'Unicode.'
        )
        log.e('please set it to a UTF-8-aware locale first,')
        log.e(
            'so as to save the video (with some Unicode characters) correctly.'
        )
        log.e('you can do it like this:')
        log.e('    (Windows)    % chcp 65001 ')
        log.e('    (Linux)      $ LC_CTYPE=en_US.UTF-8')
        sys.exit(1)
    except Exception:
        if not args.debug:
            log.e('[error] oops, something went wrong.')
            log.e(
                'don\'t panic, c\'est la vie. please try the following steps:'
            )
            log.e('  (1) Rule out any network problem.')
            log.e('  (2) Make sure lulu is up-to-date.')
            log.e('  (3) Check if the issue is already known, on')
            log.e('        https://github.com/iawia002/Lulu/issues')
            log.e('  (4) Run the command with \'--debug\' option,')
            log.e('      and report this issue with the full output.')
        else:
            print_version()
            log.i(args)
            raise
        sys.exit(1)


def google_search(url):
    keywords = match1(url, r'https?://(.*)')
    url = 'https://www.google.com/search?tbm=vid&q=%s' % parse.quote(keywords)
    page = get_content(url)
    videos = re.findall(
        r'<a href="(https?://[^"]+)" onmousedown="[^"]+">([^<]+)<', page
    )
    vdurs = re.findall(r'<span class="vdur _dwc">([^<]+)<', page)
    durs = [match1(unescape(dur), r'(\d+:\d+)') for dur in vdurs]
    print('Google Videos search:')
    for v in zip(videos, durs):
        print('- video:  {} [{}]'.format(
            unescape(v[0][1]),
            v[1] if v[1] else '?'
        ))
        print('# lulu %s' % log.sprint(v[0][0], log.UNDERLINE))
        print()
    print('Best matched result:')
    return(videos[0][0])


def url_to_module(url):
    try:
        video_host = match1(url, r'https?://([^/]+)/')
        video_url = match1(url, r'https?://[^/]+(.*)')
        assert video_host and video_url
    except AssertionError:
        url = google_search(url)
        video_host = match1(url, r'https?://([^/]+)/')
        video_url = match1(url, r'https?://[^/]+(.*)')

    if video_host.endswith('.com.cn') or video_host.endswith('.ac.cn'):
        video_host = video_host[:-3]
    domain = match1(video_host, r'(\.[^.]+\.[^.]+)$') or video_host
    assert domain, 'unsupported url: ' + url

    k = match1(domain, r'([^.]+)')
    if k in config.SITES:
        return (
            import_module('.'.join(['lulu', 'extractors', config.SITES[k]])),
            url
        )
    else:
        import http.client
        video_host = match1(url, r'https?://([^/]+)/')  # .cn could be removed
        if url.startswith('https://'):
            conn = http.client.HTTPSConnection(video_host)
        else:
            conn = http.client.HTTPConnection(video_host)
        conn.request('HEAD', video_url, headers=config.FAKE_HEADERS)
        res = conn.getresponse()
        location = res.getheader('location')
        if location and location != url and not location.startswith('/'):
            return url_to_module(location)
        else:
            return import_module('lulu.extractors.universal'), url


def any_download(url, **kwargs):
    m, url = url_to_module(url)
    m.download(url, **kwargs)


def any_download_playlist(url, **kwargs):
    m, url = url_to_module(url)
    m.download_playlist(url, **kwargs)


def main(**kwargs):
    script_main(any_download, any_download_playlist, **kwargs)



#!/usr/bin/env python

import re
import json
from base64 import b64decode

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
)


__all__ = ['kugou_download']
site_info = '酷狗音乐 kugou.com'


def kugou_download_by_hash(hash_val, info_only=False, **kwargs):
    # http://www.kugou.com/yy/album/single/536957.html
    html = get_content(
        'http://www.kugou.com/yy/index.php?r=play/getdata&hash={}'.format(
            hash_val
        )
    )
    data = json.loads(html)
    url = data['data']['play_url']
    title = data['data']['audio_name']
    songtype, ext, size = url_info(url)
    print_info(site_info, title, songtype, size)
    if not info_only:
        download_urls([url], title, ext, size, **kwargs)


def kugou_download_playlist(url, info_only=False, **kwargs):
    html = get_content(url)
    pattern = re.compile('title="(.*?)".* data="(\w*)\|.*?"')
    pairs = pattern.findall(html)
    for title, hash_val in pairs:
        kugou_download_by_hash(hash_val, info_only, **kwargs)


def kugou_download(url, info_only=False, **kwargs):
    html = get_content(url)
    if url.lower().find('5sing') != -1:
        # for 5sing.kugou.com
        ticket = match1(html, r'"ticket":\s*"(.*)"')
        j = json.loads(str(b64decode(ticket), encoding='utf-8'))
        url = j['file']
        title = j['songName']
        songtype, ext, size = url_info(url)
        print_info(site_info, title, songtype, size)
        if not info_only:
            download_urls([url], title, ext, size, **kwargs)
    else:
        # for the www.kugou.com
        hash_val = match1(url, r'hash=(\w+)')
        kugou_download_by_hash(hash_val, info_only, **kwargs)


download = kugou_download
download_playlist = kugou_download_playlist



#!/usr/bin/env python

import json
import base64
from urllib.parse import urlparse

from lulu.common import (
    print_info,
    get_content,
    download_urls,
    download_url_ffmpeg,
    general_m3u8_extractor,
    playlist_not_supported,
)


__all__ = ['zhanqi_download']
site_info = '战旗 zhanqi.tv'


def zhanqi_download(
    url, output_dir='.', merge=True, info_only=False, **kwargs
):
    path = urlparse(url).path[1:]

    if not (path.startswith('videos') or path.startswith('v2/videos')):
        path_list = path.split('/')
        room_id = path_list[1] if path_list[0] == 'topic' else path_list[0]
        zhanqi_live(
            room_id, merge=merge, output_dir=output_dir, info_only=info_only,
            **kwargs
        )
    else:
        # https://www.zhanqi.tv/videos/Lyingman/2017/01/182308.html
        # https://www.zhanqi.tv/v2/videos/215593.html
        video_id = path.split('.')[0].split('/')[-1]
        zhanqi_video(
            video_id, merge=merge, output_dir=output_dir, info_only=info_only,
            **kwargs
        )


def zhanqi_live(
    room_id, merge=True, output_dir='.', info_only=False, **kwargs
):
    api_url = (
        'https://www.zhanqi.tv/api/static/v2.2/room/domain/{}.json'.format(
            room_id
        )
    )
    json_data = json.loads(get_content(api_url))['data']
    status = json_data['status']
    if status != '4':
        raise Exception('The live stream is not online!')

    nickname = json_data['nickname']
    title = '{}：{}'.format(
        nickname, json_data['title']
    )
    video_levels = base64.b64decode(
        json_data['flashvars']['VideoLevels']
    ).decode('utf8')
    m3u8_url = json.loads(video_levels)['streamUrl']

    print_info(
        site_info, title, 'm3u8', 0, m3u8_url=m3u8_url, m3u8_type='master'
    )
    if not info_only:
        download_url_ffmpeg(
            m3u8_url, title, 'mp4', output_dir=output_dir, merge=merge
        )


def zhanqi_video(
    video_id, output_dir='.', info_only=False, merge=True, **kwargs
):
    api_url = 'https://www.zhanqi.tv/api/static/v2.2/video/{}.json'.format(
        video_id
    )
    json_data = json.loads(get_content(api_url))['data']
    title = json_data['title']
    m3u8_url = json_data['playUrl']
    urls = general_m3u8_extractor(m3u8_url)
    print_info(site_info, title, 'm3u8', 0)
    if not info_only:
        download_urls(
            urls, title, 'ts', 0, output_dir=output_dir, merge=merge, **kwargs
        )


download = zhanqi_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json
import urllib.parse
import urllib.request

from lulu.util import log
from lulu.common import (
    url_size,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['naver_download_by_url']
site_info = 'NAVER naver.com'


def naver_download_by_url(url, info_only=False, **kwargs):
    ep = 'https://apis.naver.com/rmcnmv/rmcnmv/vod/play/v2.0/{}?key={}'
    page = get_content(url)
    og_video_url = re.search(
        r"<meta\s+property=\"og:video:url\"\s+content='(.+?)'>", page
    ).group(1)
    params_dict = urllib.parse.parse_qs(
        urllib.parse.urlparse(og_video_url).query
    )
    vid = params_dict['vid'][0]
    key = params_dict['outKey'][0]
    meta_str = get_content(ep.format(vid, key))
    meta_json = json.loads(meta_str)
    if 'errorCode' in meta_json:
        log.wtf(meta_json['errorCode'])
    title = meta_json['meta']['subject']
    videos = meta_json['videos']['list']
    video_list = sorted(
        videos, key=lambda video: video['encodingOption']['width']
    )
    video_url = video_list[-1]['source']
    size = url_size(video_url)
    print_info(site_info, title, 'mp4', size)
    if not info_only:
        download_urls([video_url], title, 'mp4', size, **kwargs)


download = naver_download_by_url
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json
from urllib.parse import urlparse

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['yixia_download']
site_info = '秒拍 miaopai.com'


def yixia_miaopai_download_by_scid(scid, info_only=False, **kwargs):
    api_endpoint = (
        'http://api.miaopai.com/m/v2_channel.json?fillType=259&scid='
        '{scid}&vend=miaopai'.format(scid=scid)
    )
    html = get_content(api_endpoint)
    api_content = json.loads(html)
    video_url = match1(api_content['result']['stream']['base'], r'(.+)\?vend')
    title = api_content['result']['ext']['t']
    _type, ext, size = url_info(video_url)
    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([video_url], title, ext, size, **kwargs)


def yixia_xiaokaxiu_download_by_scid(scid, info_only=False, **kwargs):
    site_info = '小咖秀 xiaokaxiu.com'
    api_endpoint = (
        'http://api.xiaokaxiu.com/video/web/get_play_video?'
        'scid={scid}'.format(scid=scid)
    )
    html = get_content(api_endpoint)
    api_content = json.loads(html)
    video_url = api_content['data']['linkurl']
    title = api_content['data']['title']
    _type, ext, size = url_info(video_url)
    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([video_url], title, ext, size, **kwargs)


def yixia_download(url, info_only=False, **kwargs):
    hostname = urlparse(url).hostname
    if 'miaopai.com' in hostname:  # Miaopai
        yixia_download_by_scid = yixia_miaopai_download_by_scid

        scid = match1(url, r'miaopai\.com/show/channel/(.+)\.htm') or \
            match1(url, r'miaopai\.com/show/(.+)\.htm') or \
            match1(url, r'm\.miaopai\.com/show/channel/(.+)\.htm') or \
            match1(url, r'm\.miaopai\.com/show/channel/([^\?]+)')

    elif 'xiaokaxiu.com' in hostname:  # Xiaokaxiu
        yixia_download_by_scid = yixia_xiaokaxiu_download_by_scid

        scid = match1(url, r'v.xiaokaxiu.com/v/(.+)\.html') or \
            match1(url, r'm.xiaokaxiu.com/m/(.+)\.html')

    else:
        pass

    yixia_download_by_scid(scid, info_only, **kwargs)


download = yixia_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json
import time
import base64
import random
import urllib
import hashlib

from lulu.common import (
    match1,
    url_info,
    urls_size,
    print_info,
    get_content,
    download_urls,
    url_locations,
    playlist_not_supported,
)


__all__ = ['letv_download', 'letvcloud_download', 'letvcloud_download_by_vu']
site_info = '乐视 le.com'


def calcTimeKey(t):
    def ror(val, r_bits):
        return ((val & (2**32-1)) >> r_bits % 32) | (
            val << (32-(r_bits % 32)) & (2**32-1)
        )
    magic = 185025305
    return ror(t, magic % 17) ^ magic


def decode(data):
    version = data[0:5]
    if version.lower() == b'vc_01':
        # get real m3u8
        loc2 = data[5:]
        length = len(loc2)
        loc4 = [0]*(2*length)
        for i in range(length):
            loc4[2*i] = loc2[i] >> 4
            loc4[2*i+1] = loc2[i] & 15
        loc6 = loc4[len(loc4)-11:]+loc4[:len(loc4)-11]
        loc7 = [0]*length
        for i in range(length):
            loc7[i] = (loc6[2 * i] << 4) + loc6[2*i+1]
        return ''.join([chr(i) for i in loc7])
    else:
        # directly return
        return data


def video_info(vid, **kwargs):
    url = (
        'http://player-pc.le.com/mms/out/video/playJson?id={}&platid=1&'
        'splatid=101&format=1&tkey={}&domain=www.le.com&region=cn&source='
        '1000&accesyx=1'.format(vid, calcTimeKey(int(time.time())))
    )
    r = get_content(url, decoded=False)
    info = json.loads(str(r, 'utf-8'))
    info = info['msgs']
    stream_id = None
    support_stream_id = info['playurl']['dispatch'].keys()
    if 'stream_id' in kwargs and \
            kwargs['stream_id'].lower() in support_stream_id:
        stream_id = kwargs['stream_id']
    else:
        if '1080p' in support_stream_id:
            stream_id = '1080p'
        elif '720p' in support_stream_id:
            stream_id = '720p'
        else:
            stream_id = sorted(
                support_stream_id, key=lambda i: int(i[1:])
            )[-1]

    url = info['playurl']['domain'][0] \
        + info['playurl']['dispatch'][stream_id][0]
    uuid = hashlib.sha1(url.encode('utf8')).hexdigest() + '_0'
    ext = info['playurl']['dispatch'][stream_id][1].split('.')[-1]
    url = url.replace('tss=0', 'tss=ios')
    url += (
        '&m3v=1&termid=1&format=1&hwtype=un&ostype=MacOS10.12.4&p1=1&p2=10&p3'
        '=-&expect=3&tn={}&vid={}&uuid={}&sign=letv'.format(
            random.random(), vid, uuid
        )
    )
    r2 = get_content(url, decoded=False)
    info2 = json.loads(str(r2, 'utf-8'))

    # hold on ! more things to do
    # to decode m3u8 (encoded)
    suffix = '&r={}&appid=500'.format(str(int(time.time() * 1000)))
    m3u8 = get_content(info2['location']+suffix, decoded=False)
    m3u8_list = decode(m3u8)
    urls = re.findall(r'^[^#][^\r]*', m3u8_list, re.MULTILINE)
    return ext, urls


def letv_download_by_vid(vid, title, info_only=False, **kwargs):
    ext, urls = video_info(vid, **kwargs)
    size = 0
    for i in urls:
        _, _, tmp = url_info(i)
        size += tmp

    print_info(site_info, title, ext, size)
    if not info_only:
        download_urls(urls, title, ext, size, **kwargs)


def letvcloud_download_by_vu(vu, uu, title=None, info_only=False, **kwargs):
    argumet_dict = {
        'cf': 'flash', 'format': 'json', 'ran': str(int(time.time())),
        'uu': str(uu), 'ver': '2.2', 'vu': str(vu),
    }
    # ALL YOUR BASE ARE BELONG TO US
    sign_key = '2f9d6924b33a165a6d8b5d3d42f4f987'
    str2Hash = ''.join(
        [i + argumet_dict[i] for i in sorted(argumet_dict)]
    ) + sign_key
    sign = hashlib.md5(str2Hash.encode('utf-8')).hexdigest()
    request_info = urllib.request.Request(
        'http://api.letvcloud.com/gpc.php?{}&sign={}'.format(
            '&'.join(
                ['{}={}'.format(i, argumet_dict[i]) for i in argumet_dict]
            ),
            sign
        )
    )
    response = urllib.request.urlopen(request_info)
    data = response.read()
    info = json.loads(data.decode('utf-8'))
    type_available = []
    for video_type in info['data']['video_info']['media']:
        type_available.append({
            'video_url': info['data']['video_info']['media'][video_type][
                'play_url'
            ]['main_url'],
            'video_quality': int(
                info['data']['video_info']['media'][video_type][
                    'play_url'
                ]['vtype']
            )
        })
    urls = [base64.b64decode(sorted(
        type_available,
        key=lambda x: x['video_quality'])[-1]['video_url']
    ).decode('utf-8')]
    size = urls_size(urls)
    ext = 'mp4'
    print_info(site_info, title, ext, size)
    if not info_only:
        download_urls(urls, title, ext, size, **kwargs)


def letvcloud_download(url, info_only=False, **kwargs):
    qs = urllib.parse.urlparse(url).query
    vu = match1(qs, r'vu=([\w]+)')
    uu = match1(qs, r'uu=([\w]+)')
    title = 'LETV-{}'.format(vu)
    letvcloud_download_by_vu(
        vu, uu, title=title, info_only=info_only, **kwargs
    )


def letv_download(url, info_only=False, **kwargs):
    url = url_locations([url])[0]
    if re.match(r'http://yuntv.letv.com/', url):
        letvcloud_download(url, info_only=info_only, **kwargs)
    elif 'sports.le.com' in url:
        html = get_content(url)
        vid = match1(url, r'video/(\d+)\.html')
        title = match1(html, r'<h2 class="title">([^<]+)</h2>')
        letv_download_by_vid(vid, title=title, info_only=info_only, **kwargs)
    else:
        html = get_content(url)
        vid = match1(url, r'http://www.letv.com/ptv/vplay/(\d+).html') or \
            match1(url, r'http://www.le.com/ptv/vplay/(\d+).html') or \
            match1(html, r'vid="(\d+)"')
        title = match1(html, r'name="irTitle" content="(.*?)"')
        letv_download_by_vid(vid, title=title, info_only=info_only, **kwargs)


download = letv_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json

from lulu.common import (
    match1,
    get_content,
    playlist_not_supported,
)
from lulu.extractor import VideoExtractor


__all__ = ['cntv_download', 'cntv_download_by_id']
site_info = '央视网 cctv.com'


class CNTV(VideoExtractor):
    name = site_info
    stream_types = [
        {
            'id': '1', 'video_profile': '1280x720_2000kb/s',
            'map_to': 'chapters4'
        },
        {
            'id': '2', 'video_profile': '1280x720_1200kb/s',
            'map_to': 'chapters3'
        },
        {
            'id': '3', 'video_profile': '640x360_850kb/s',
            'map_to': 'chapters2'
        },
        {
            'id': '4', 'video_profile': '480x270_450kb/s',
            'map_to': 'chapters'
        },
        {
            'id': '5', 'video_profile': '320x180_200kb/s',
            'map_to': 'lowChapters'
        },
    ]

    ep = 'http://vdn.apps.cntv.cn/api/getHttpVideoInfo.do?pid={}'

    def __init__(self):
        super().__init__()
        self.api_data = None

    def prepare(self, **kwargs):
        self.api_data = json.loads(
            get_content(self.__class__.ep.format(self.vid))
        )
        self.title = self.api_data['title']
        for s in self.api_data['video']:
            for st in self.__class__.stream_types:
                if st['map_to'] == s:
                    urls = self.api_data['video'][s]
                    src = [u['url'] for u in urls]
                    stream_data = dict(
                        src=src, size=0, container='mp4',
                        video_profile=st['video_profile']
                    )
                    self.streams[st['id']] = stream_data


def cntv_download_by_id(rid, **kwargs):
    CNTV().download_by_vid(rid, **kwargs)


def cntv_download(url, **kwargs):
    page = get_content(url)
    if re.match(r'http://tv\.cntv\.cn/video/(\w+)/(\w+)', url):
        rid = match1(url, r'http://tv\.cntv\.cn/video/\w+/(\w+)')
    elif re.match(r'http://tv\.cctv\.com/\d+/\d+/\d+/\w+.shtml', url):
        rid = match1(page, r'var guid = "(\w+)"')
    elif re.match(
        r'http://\w+\.cntv\.cn/(\w+/\w+/(classpage/video/)?)?\d+/\d+\.shtml',
        url
    ) or re.match(r'http://\w+.cntv.cn/(\w+/)*VIDE\d+.shtml', url) or re.match(
        r'http://(\w+).cntv.cn/(\w+)/classpage/video/(\d+)/(\d+).shtml', url
    ) or re.match(
        r'http://\w+.cctv.com/\d+/\d+/\d+/\w+.shtml', url
    ) or re.match(r'http://\w+.cntv.cn/\d+/\d+/\d+/\w+.shtml', url):
        rid = match1(page, r'videoCenterId","(\w+)"')
        if rid is None:
            guid = re.search(r'guid\s*=\s*"([0-9a-z]+)"', page).group(1)
            rid = guid
    elif re.match(r'http://xiyou.cntv.cn/v-[\w-]+\.html', url):
        rid = match1(url, r'http://xiyou.cntv.cn/v-([\w-]+)\.html')
    else:
        raise NotImplementedError(url)

    CNTV().download_by_vid(rid, **kwargs)


download = cntv_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json
from urllib import parse

from lulu import config
from lulu.common import (
    match1,
    url_size,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.extractors.yixia import yixia_miaopai_download_by_scid


__all__ = ['weibo_download']
site_info = '微博 weibo.com'


def weibo_download_by_fid(fid, info_only=False, **kwargs):
    page_url = 'http://video.weibo.com/show?fid={}&type=mp4'.format(fid)

    mobile_page = get_content(page_url, headers=config.FAKE_HEADERS_MOBILE)
    url = match1(mobile_page, r'<video id=.*?src=[\'"](.*?)[\'"]\W')
    title = match1(mobile_page, r'<title>((.|\n)+?)</title>')
    if not title:
        title = fid
    title = title.replace('\n', '_')
    ext, size = 'mp4', url_size(url)
    print_info(site_info, title, ext, size)
    if not info_only:
        download_urls([url], title, ext, size, **kwargs)


def get_fid(url):
    return match1(url, r'\?fid=(\d{4}:\w{32})')


def weibo_download(url, info_only=False, **kwargs):
    fid = get_fid(url)
    if fid:
        weibo_download_by_fid(fid, info_only, **kwargs)
    elif '/p/230444' in url:
        fid = match1(url, r'/p/230444(\w+)')
        weibo_download_by_fid('1034:{}'.format(fid), info_only, **kwargs)
    else:
        status_id = url.split('?')[0].split('/')[-1]
        video_info = json.loads(
            get_content(
                'https://m.weibo.cn/statuses/show?id={}'.format(status_id),
                headers=config.FAKE_HEADERS_MOBILE
            )
        )
        video_url = video_info['data']['page_info']['media_info'][
            'stream_url_hd'
        ]
        if not video_url:
            video_url = parse.unquote(
                video_info['data']['page_info']['page_url']
            )
            if 'miaopai' in video_url:
                # https://weibo.cn/sinaurl/blocked817bc30b?u=http%3A%2F%2Fmiaopai.com%2Fshow%2FnPWJvdR4z2Bg1Sz3PJpNYffjpDgEiuv4msALgw__.htm  # noqa
                scid = match1(
                    video_url.split('=')[-1],
                    r'.*?miaopai.com/show/(.+)\.htm'
                )
                return yixia_miaopai_download_by_scid(
                    scid, info_only=info_only, **kwargs
                )
            elif 'fid=' in video_url:
                # http://video.weibo.com/show?fid=1034:b91d1ecf44b0e2f18c436d819744b333  # noqa
                fid = get_fid(video_url)
                return weibo_download_by_fid(fid, info_only, **kwargs)
        title = video_info['data']['page_info']['content2']
        video_format = 'mp4'
        size = url_size(video_url)
        print_info(
            site_info=site_info, title=title, type=video_format, size=size
        )
        if not info_only:
            download_urls(
                urls=[video_url], title=title, ext=video_format,
                total_size=size, **kwargs
            )


download = weibo_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json

from lulu.common import (
    match1,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.extractors.sohu import sohu_download


__all__ = ['w56_download', 'w56_download_by_id']
site_info = '56网 56.com'


def w56_download_by_id(_id, title=None, info_only=False, **kwargs):
    content = json.loads(get_content(
        'http://vxml.56.com/json/{}/?src=site'.format(_id)
    ))
    info = content['info']
    title = title or info['Subject']
    assert title
    hd = info['hd']
    assert hd in (0, 1, 2)
    hd_types = [['normal', 'qvga'], ['clear', 'vga'], ['super', 'wvga']][hd]
    files = [x for x in info['rfiles'] if x['type'] in hd_types]
    assert len(files) == 1
    size = int(files[0]['filesize'])
    url = files[0]['url'] + '&prod=56'
    ext = 'mp4'

    print_info(site_info, title, ext, size)
    if not info_only:
        download_urls([url], title, ext, size, **kwargs)


def w56_download(url, info_only=False, **kwargs):
    content = get_content(url)
    sohu_url = match1(content, r'url:\s*"(.+)"')
    if sohu_url:
        sohu_download(sohu_url, info_only=info_only, **kwargs)
        return
    _id = match1(url, r'http://www.56.com/u\d+/v_(\w+).html') or \
        match1(url, r'http://www.56.com/.*vid-(\w+).html')
    w56_download_by_id(_id, info_only=info_only, **kwargs)


download = w56_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['veoh_download']
site_info = 'Veoh veoh.com'


def veoh_download(url, info_only=False, **kwargs):
    '''Get item_id'''
    if re.match(r'http://www.veoh.com/watch/\w+', url):
        item_id = match1(url, r'http://www.veoh.com/watch/(\w+)')
    elif re.match(r'http://www.veoh.com/m/watch.php\?v=\.*', url):
        item_id = match1(url, r'http://www.veoh.com/m/watch.php\?v=(\w+)')
    else:
        raise NotImplementedError('Cannot find item ID')
    veoh_download_by_id(item_id, info_only=info_only, **kwargs)


def veoh_download_by_id(item_id, info_only=False, **kwargs):
    """Source: Android mobile
    """
    webpage_url = (
        'http://www.veoh.com/m/watch.php?v={item_id}&quality=1'.format(
            item_id=item_id
        )
    )
    # grab download URL
    a = get_content(webpage_url, decoded=True)
    url = match1(a, r'<source src="(.*?)\"\W')

    # grab title
    title = match1(a, r'<meta property="og:title" content="([^"]*)"')

    type_, ext, size = url_info(url)
    print_info(site_info, title, type_, size)
    if not info_only:
        download_urls([url], title, ext, total_size=None, **kwargs)


download = veoh_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['magisto_download']
site_info = 'Magisto magisto.com'


def magisto_download(url, info_only=False, **kwargs):
    video_hash = match1(url, r'video\/([a-zA-Z0-9]+)')
    api_url = 'https://www.magisto.com/api/video/{}'.format(video_hash)
    content = get_content(api_url)
    data = json.loads(content)
    title1 = data['title']
    title2 = data['creator']
    title = '{} - {}'.format(title1, title2)
    url = data['video_direct_url']
    _type, ext, size = url_info(url)

    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([url], title, ext, size, **kwargs)


download = magisto_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['iqilu_download']
site_info = '齐鲁网 iqilu.com'


def iqilu_download(url, info_only=False, **kwargs):
    if re.match(r'http://v.iqilu.com/\w+', url):
        patt = r'url\s*:\s*\[([^\]]+)\]'

        # URL in webpage
        html = get_content(url)
        player_data = '[{}]'.format(match1(html, patt))
        urls = json.loads(player_data)
        url = urls[0]['stream_url']

        # grab title
        title = match1(html, r'<meta name="description" content="(.*?)\"\W')

        _type, ext, size = url_info(url)
        print_info(site_info, title, _type, size)
        if not info_only:
            download_urls([url], title, ext, size, **kwargs)


download = iqilu_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json
import time
from random import random
from urllib.parse import urlparse

from lulu.common import (
    match1,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)

__all__ = ['sohu_download']
site_info = '搜狐视频 tv.sohu.com'


'''
Changelog:
    1. http://tv.sohu.com/upload/swf/20150604/Main.swf
        new api
'''


def real_url(host, vid, tvid, new, clipURL, ck):
    url = (
        'http://{}/?prot=9&prod=flash&pt=1&file={}&new={}&key={}&vid={}'
        '&uid={}&t={}&rb=1'.format(
            host, clipURL, new, ck, str(vid), str(int(time.time()*1000)),
            str(random())
        )
    )
    return json.loads(get_content(url))['url']


def sohu_download(url, info_only=False, **kwargs):
    if re.match(r'http://share.vrs.sohu.com', url):
        vid = match1(url, 'id=(\d+)')
    else:
        html = get_content(url)
        vid = match1(html, r'\Wvid\s*[\:=]\s*[\'"]?(\d+)[\'"]?')
    assert vid

    if re.match(r'http[s]://tv.sohu.com/', url):
        info = json.loads(get_content(
            'http://hot.vrs.sohu.com/vrs_flash.action?vid={}'.format(vid)
        ))
        for qtyp in ['oriVid', 'superVid', 'highVid', 'norVid', 'relativeId']:
            if 'data' in info:
                hqvid = info['data'][qtyp]
            else:
                hqvid = info[qtyp]
            if hqvid != 0 and hqvid != vid:
                info = json.loads(get_content(
                    'http://hot.vrs.sohu.com/vrs_flash.action?vid={}'.format(
                        hqvid
                    )
                ))
                if 'allot' not in info:
                    continue
                break
        host = info['allot']
        tvid = info['tvid']
        urls = []
        data = info['data']
        title = data['tvName']
        size = sum(data['clipsBytes'])
        assert len(data['clipsURL']) == len(data['clipsBytes']) \
            == len(data['su'])
        for new, clip, ck in zip(data['su'], data['clipsURL'], data['ck']):
            clipURL = urlparse(clip).path
            urls.append(real_url(host, hqvid, tvid, new, clipURL, ck))

    else:
        info = json.loads(get_content(
            'http://my.tv.sohu.com/play/videonew.do?vid={}&referer='
            'http://my.tv.sohu.com'.format(vid)
        ))
        host = info['allot']
        tvid = info['tvid']
        urls = []
        data = info['data']
        title = data['tvName']
        size = sum(map(int, data['clipsBytes']))
        assert len(data['clipsURL']) == len(data['clipsBytes']) \
            == len(data['su'])
        for new, clip, ck, in zip(data['su'], data['clipsURL'], data['ck']):
            clipURL = urlparse(clip).path
            urls.append(real_url(host, vid, tvid, new, clipURL, ck))

    print_info(site_info, title, 'mp4', size)
    if not info_only:
        download_urls(urls, title, 'mp4', size, refer=url, **kwargs)


download = sohu_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import urllib

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['suntv_download']
site_info = 'ISUNVIDEO isuntv.com'


def suntv_download(url, info_only=False, **kwargs):
    if re.match(r'http://www.isuntv.com/\w+', url):
        API_URL = (
            'http://www.isuntv.com/ajaxpro/SunTv.pro_vod_playcatemp4,'
            'App_Web_playcatemp4.ascx.9f08f04f.ashx'
        )
        itemid = match1(url, r'http://www.isuntv.com/pro/ct(\d+).html')
        values = {'itemid': itemid, 'vodid': ''}

        data = str(values).replace("'", '"')
        data = data.encode('utf-8')
        req = urllib.request.Request(API_URL, data)
        req.add_header('AjaxPro-Method', 'ToPlay')  # important!
        resp = urllib.request.urlopen(req)
        respData = resp.read()
        respData = respData.decode('ascii').strip('"')  # Ahhhhhhh!

        video_url = 'http://www.isuntv.com' + str(respData)

        html = get_content(url, decoded=False)
        html = html.decode('gbk')
        title = match1(html, '<title>([^<]+)').strip()  # get rid of \r\n s

        size = 0
        _type, ext, size = url_info(video_url)

        print_info(site_info, title, _type, size)
        if not info_only:
            download_urls([url], title, 'mp4', size, **kwargs)


download = suntv_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json
import random
from urllib.parse import (
    urlparse,
    parse_qs,
)

from lulu.config import FAKE_HEADERS
from lulu.common import (
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['fantasy_download']
site_info = 'fantasy.tv'


def fantasy_download_by_id_channelId(
    id=0, channelId=0, output_dir='.', merge=True, info_only=False, **kwargs
):
    api_url = (
        'http://www.fantasy.tv/tv/playDetails.action?'
        'myChannelId=1&id={id}&channelId={channelId}&t={t}'.format(
            id=id,
            channelId=channelId,
            t=str(random.random())
        )
    )
    html = get_content(api_url)
    html = json.loads(html)

    if int(html['status']) != 100000:
        raise Exception('API error!')

    title = html['data']['tv']['title']

    video_url = html['data']['tv']['videoPath']
    headers = FAKE_HEADERS.copy()
    headers['Referer'] = api_url
    type, ext, size = url_info(video_url, headers=headers)

    print_info(site_info, title, type, size)
    if not info_only:
        download_urls(
            [video_url], title, ext, size, output_dir, merge=merge,
            headers=headers
        )


def fantasy_download(
    url, output_dir='.', merge=True, info_only=False, **kwargs
):
    if 'fantasy.tv' not in url:
        raise Exception('Wrong place!')

    q = parse_qs(urlparse(url).query)

    if 'tvId' not in q or 'channelId' not in q:
        raise Exception('No enough arguments!')

    tvId = q['tvId'][0]
    channelId = q['channelId'][0]

    fantasy_download_by_id_channelId(
        id=tvId, channelId=channelId, output_dir=output_dir, merge=merge,
        info_only=info_only, **kwargs
    )


download = fantasy_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import xml.etree.ElementTree as ET

from lulu.common import (
    url_info,
    get_content,
)
from lulu.util import log
from lulu.extractor import VideoExtractor


class BokeCC(VideoExtractor):
    name = 'BokeCC'

    # we do now know for now, as we have to check the output from the API
    stream_types = []

    API_ENDPOINT = 'http://p.bokecc.com/'

    def download_by_id(
        self, vid='', title=None, output_dir='.', merge=True, info_only=False,
        **kwargs
    ):
        """self, str->None

        Keyword arguments:
        self: self
        vid: The video ID for BokeCC cloud, something like
        FE3BB999594978049C33DC5901307461

        Calls the prepare() to download the video.

        If no title is provided, this method shall try to find a proper title
        with the information providin within the
        returned content of the API."""

        assert vid

        self.prepare(vid=vid, title=title, **kwargs)

        self.extract(**kwargs)

        self.download(
            output_dir=output_dir, merge=merge, info_only=info_only,
            **kwargs
        )

    def prepare(self, vid='', title=None, **kwargs):
        assert vid

        api_url = self.API_ENDPOINT + \
            'servlet/playinfo?vid={vid}&m=0'.format(vid=vid)  # return XML

        html = get_content(api_url)
        self.tree = ET.ElementTree(ET.fromstring(html))

        if self.tree.find('result').text != '1':
            log.wtf('API result says failed!')
            raise

        if title is None:
            self.title = '_'.join(
                [i.text for i in self.tree.iterfind(
                    'video/videomarks/videomark/markdesc'
                )]
            )
        else:
            self.title = title

        if not title:
            self.title = vid

        for i in self.tree.iterfind('video/quality'):
            quality = i.attrib['value']
            url = i[0].attrib['playurl']
            self.stream_types.append({
                'id': quality, 'video_profile': i.attrib['desp']
            })
            self.streams[quality] = {
                'url': url, 'video_profile': i.attrib['desp']
            }
            self.streams_sorted = [
                dict(
                    [('id', stream_type['id'])] +
                    list(self.streams[stream_type['id']].items())
                ) for stream_type in self.__class__.stream_types
                if stream_type['id'] in self.streams
            ]

    def extract(self, **kwargs):
        for i in self.streams:
            s = self.streams[i]
            _, s['container'], s['size'] = url_info(s['url'])
            s['src'] = [s['url']]
        if 'stream_id' in kwargs and kwargs['stream_id']:
            # Extract the stream
            stream_id = kwargs['stream_id']

            if stream_id not in self.streams:
                log.e('[Error] Invalid video format.')
                log.e(
                    'Run \'-i\' command with no specific video format to '
                    'view all available formats.'
                )
                exit(2)
        else:
            # Extract stream with the best quality
            stream_id = self.streams_sorted[0]['id']
            _, s['container'], s['size'] = url_info(s['url'])
            s['src'] = [s['url']]


site = BokeCC()

# I don't know how to call the player directly so I just put it here
# just in case anyone touchs it -- Beining@Aug.24.2016
# download = site.download_by_url
# download_playlist = site.download_by_url

bokecc_download_by_id = site.download_by_id



#!/usr/bin/env python

import re
from html import unescape

from lulu.common import (
    match1,
    print_info,
    get_content,
    download_rtmp_url,
    playlist_not_supported,
)


site_info = 'thePlatform.com'


def theplatform_download_by_pid(
    pid, title, output_dir='.', merge=True, info_only=False, **kwargs
):
    smil_url = (
        'http://link.theplatform.com/s/dJ5BDC/{}/meta.smil?format=smil'
        '&mbr=true'.format(pid)
    )
    smil = get_content(smil_url)
    smil_base = unescape(match1(smil, r'<meta base="([^"]+)"'))
    smil_videos = {
        y: x for x, y in dict(
            re.findall(r'<video src="([^"]+)".+height="([^"]+)"', smil)
        ).items()
    }
    for height in ['1080', '720', '480', '360', '240', '216']:
        if height in smil_videos:
            smil_video = smil_videos[height]
            break
    assert smil_video

    _type, ext, size = 'mp4', 'mp4', 0

    print_info(site_info, title, _type, size)
    if not info_only:
        download_rtmp_url(
            url=smil_base, title=title, ext=ext,
            params={"-y": '{}:{}'.format(ext, smil_video)},
            output_dir=output_dir
        )


download = theplatform_download_by_pid
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import string
from urllib import parse

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['ku6_download']
site_info = '酷6网 ku6.com'


def ku6_download(url, info_only=False, **kwargs):
    page = get_content(url)
    video = match1(page, r'type: "video/mp4", src: "(.+)"').replace(' ', '%20')
    video = parse.quote(
        video, safe=string.printable
    )
    title = match1(page, r'document.title = "(.+)"')
    _type, ext, size = url_info(video)
    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([video], title, ext, size, **kwargs)


download = ku6_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

from lulu.common import (
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.util.parser import get_parser


__all__ = ['giphy_download']
site_info = 'GIPHY giphy.com'


def giphy_download(url, info_only=False, **kwargs):
    html = get_content(url)
    parser = get_parser(html)

    title = parser.find('meta', property='og:title')['content']
    gif = parser.find('meta', property='og:image')['content']
    video = parser.find('meta', property='og:video')['content']

    for url in [gif, video]:
        _type, ext, size = url_info(url)
        print_info(site_info, title, _type, size)
        if not info_only:
            download_urls([url], title, ext, size, **kwargs)


download = giphy_download
download_playlist = playlist_not_supported(site_info)



# coding=utf-8

import re
from urllib import parse
from xml.dom.minidom import parseString

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    get_filename,
    download_urls,
    playlist_not_supported,
)
from lulu.util import log


__all__ = ['xiami_download']
site_info = '虾米音乐 xiami.com'


def location_dec(str):
    head = int(str[0])
    str = str[1:]
    rows = head
    cols = int(len(str)/rows) + 1

    out = ''
    full_row = len(str) % head
    for c in range(cols):
        for r in range(rows):
            if c == (cols - 1) and r >= full_row:
                continue
            if r < full_row:
                char = str[r*cols+c]
            else:
                char = str[cols*full_row+(r-full_row)*(cols-1)+c]
            out += char
    return parse.unquote(out).replace('^', '0')


def xiami_download_lyric(lrc_url, file_name, output_dir):
    lrc = get_content(lrc_url)
    filename = get_filename(file_name)
    if len(lrc) > 0:
        with open(
            '{}/{}.lrc'.format(output_dir, filename), 'w', encoding='utf-8'
        ) as x:
            x.write(lrc)


def xiami_download_pic(pic_url, file_name, output_dir):
    pic_url = pic_url.replace('_1', '')
    pos = pic_url.rfind('.')
    ext = pic_url[pos:]
    pic = get_content(pic_url, decoded=False)
    if len(pic) > 0:
        with open(
            '{}/{}{}'.format(output_dir, file_name.replace('/', '-'), ext),
            'wb'
        ) as x:
            x.write(pic)


def xiami_download_song(sid, output_dir='.', info_only=False):
    xml = get_content(
        'http://www.xiami.com/song/playlist/id/{}/object_name/default/'
        'object_id/0'.format(sid)
    )
    doc = parseString(xml)
    i = doc.getElementsByTagName('track')[0]
    artist = i.getElementsByTagName('artist')[0].firstChild.nodeValue
    album_name = i.getElementsByTagName('album_name')[0].firstChild.nodeValue
    song_title = i.getElementsByTagName('name')[0].firstChild.nodeValue
    url = location_dec(
        i.getElementsByTagName('location')[0].firstChild.nodeValue
    )
    try:
        lrc_url = i.getElementsByTagName('lyric')[0].firstChild.nodeValue
    except Exception:
        pass
    type_, ext, size = url_info(url)
    if not ext:
        ext = 'mp3'

    print_info(site_info, song_title, ext, size)
    if not info_only:
        file_name = '{} - {} - {}'.format(song_title, artist, album_name)
        download_urls([url], file_name, ext, size, output_dir)
        try:
            xiami_download_lyric(lrc_url, file_name, output_dir)
        except Exception:
            pass


def xiami_download_showcollect(cid, output_dir='.', info_only=False):
    html = get_content(
        'http://www.xiami.com/song/showcollect/id/{}'.format(cid)
    )
    collect_name = match1(html, r'<title>(.*)</title>')

    xml = get_content(
        'http://www.xiami.com/song/playlist/id/{}/type/3'.format(cid)
    )
    doc = parseString(xml)
    output_dir = '{}/[{}]'.format(output_dir, collect_name)
    tracks = doc.getElementsByTagName('track')
    track_nr = 1
    for i in tracks:
        artist = album_name = song_title = url = ''
        try:
            song_id = i.getElementsByTagName('song_id')[0].firstChild.nodeValue
            artist = i.getElementsByTagName('artist')[0].firstChild.nodeValue
            album_name = i.getElementsByTagName(
                'album_name'
            )[0].firstChild.nodeValue
            song_title = i.getElementsByTagName(
                'title'
            )[0].firstChild.nodeValue
            url = location_dec(
                i.getElementsByTagName('location')[0].firstChild.nodeValue
            )
        except Exception:
            log.e(
                'Song {} failed. [Info Missing] artist: {}, album: {}, title:'
                ' {}, url: {}'.format(
                    song_id, artist, album_name, song_title, url
                )
            )
            continue
        try:
            lrc_url = i.getElementsByTagName('lyric')[0].firstChild.nodeValue
        except Exception:
            pass
        type_, ext, size = url_info(url)
        if not ext:
            ext = 'mp3'

        print_info(site_info, song_title, ext, size)
        if not info_only:
            file_name = '{:0>2d}.{} - {} - {}'.format(
                track_nr, song_title, artist, album_name
            )
            download_urls([url], file_name, ext, size, output_dir)
            try:
                xiami_download_lyric(lrc_url, file_name, output_dir)
            except Exception:
                pass

        track_nr += 1


def xiami_download_album(aid, output_dir='.', info_only=False):
    xml = get_content(
        'http://www.xiami.com/song/playlist/id/{}/type/1'.format(aid)
    )
    album_name = match1(xml, r'<album_name><!\[CDATA\[(.*)\]\]>')
    artist = match1(xml, r'<artist><!\[CDATA\[(.*)\]\]>')
    doc = parseString(xml)
    output_dir = '{}/{} - {}'.format(output_dir, artist, album_name)
    track_list = doc.getElementsByTagName('trackList')[0]
    tracks = track_list.getElementsByTagName('track')
    track_nr = 1
    pic_exist = False
    for i in tracks:
        # in this xml track tag is used for both "track in a trackList" and
        # track no dirty here
        if i.firstChild.nodeValue is not None:
            continue
        song_title = i.getElementsByTagName('songName')[0].firstChild.nodeValue
        url = location_dec(
            i.getElementsByTagName('location')[0].firstChild.nodeValue
        )
        try:
            lrc_url = i.getElementsByTagName('lyric')[0].firstChild.nodeValue
        except Exception:
            pass
        if not pic_exist:
            pic_url = i.getElementsByTagName('pic')[0].firstChild.nodeValue
        type_, ext, size = url_info(url)
        if not ext:
            ext = 'mp3'

        print_info(site_info, song_title, ext, size)
        if not info_only:
            file_name = '{:0>2d}.{}'.format(track_nr, song_title)
            download_urls([url], file_name, ext, size, output_dir)
            try:
                xiami_download_lyric(lrc_url, file_name, output_dir)
            except Exception:
                pass
            if not pic_exist:
                xiami_download_pic(pic_url, 'cover', output_dir)
                pic_exist = True

        track_nr += 1


def xiami_download_mv(url, output_dir='.', merge=True, info_only=False):
    # FIXME: broken merge
    page = get_content(url)
    title = re.findall('<title>([^<]+)', page)[0]
    vid, uid = re.findall(r'vid:"(\d+)",uid:"(\d+)"', page)[0]
    api_url = (
        'http://cloud.video.taobao.com/videoapi/info.php?vid={}&uid={}'.format(
            vid, uid
        )
    )
    result = get_content(api_url)
    doc = parseString(result)
    video_url = doc.getElementsByTagName('video_url')[-1].firstChild.nodeValue
    length = int(doc.getElementsByTagName('length')[-1].firstChild.nodeValue)

    v_urls = []
    k_start = 0
    total_size = 0
    while True:
        k_end = k_start + 20000000
        if k_end >= length:
            k_end = length - 1
        v_url = video_url + '/start_{}/end_{}/1.flv'.format(k_start, k_end)
        try:
            _, ext, size = url_info(v_url)
        except Exception:
            break
        v_urls.append(v_url)
        total_size += size
        k_start = k_end + 1

    print_info(site_info, title, ext, total_size)
    if not info_only:
        download_urls(v_urls, title, ext, total_size, output_dir, merge=merge)


def xiami_download(url, output_dir='.', merge=True, info_only=False, **kwargs):
    # albums
    if re.match(r'http://www.xiami.com/album/\d+', url):
        _id = match1(url, r'http://www.xiami.com/album/(\d+)')
        xiami_download_album(_id, output_dir, info_only)
    elif re.match(r'http://www.xiami.com/album/\w+', url):
        page = get_content(url)
        album_id = re.search(
            r'rel="canonical"\s+href="http://www.xiami.com/album/([^"]+)"',
            page
        ).group(1)
        xiami_download_album(album_id, output_dir, info_only)

    # collections
    if re.match(r'http://www.xiami.com/collect/\d+', url):
        _id = match1(url, r'http://www.xiami.com/collect/(\d+)')
        xiami_download_showcollect(_id, output_dir, info_only)

    # single track
    if re.match(r'http://www.xiami.com/song/\d+\b', url):
        _id = match1(url, r'http://www.xiami.com/song/(\d+)')
        xiami_download_song(_id, output_dir, info_only)
    elif re.match(r'http://www.xiami.com/song/\w+', url):
        html = get_content(url)
        _id = match1(
            html, r'rel="canonical" href="http://www.xiami.com/song/([^"]+)"'
        )
        xiami_download_song(_id, output_dir, info_only)

    if re.match('http://www.xiami.com/song/detail/id/\d+', url):
        _id = match1(url, r'http://www.xiami.com/song/detail/id/(\d+)')
        xiami_download_song(_id, output_dir, info_only)

    if re.match('http://www.xiami.com/mv', url):
        xiami_download_mv(url, output_dir, merge=merge, info_only=info_only)


download = xiami_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json

from lulu.common import (
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['soundcloud_download', 'soundcloud_download_by_id']
site_info = 'SoundCloud soundCloud.com'

client_id = 'WKcQQdEZw7Oi01KqtHWxeVSxNyRzgT8M'


def soundcloud_download_by_id(_id, title=None, info_only=False, **kwargs):
    assert title
    url = 'https://api.soundcloud.com/tracks/{}/{}?client_id={}'.format(
        _id, 'stream', client_id
    )

    _type, ext, size = url_info(url)
    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([url], title, ext, size, **kwargs)


def soundcloud_i1_api(track_id):
    url = (
        'https://api.soundcloud.com/i1/tracks/{}/streams?client_id={}'.format(
            track_id, client_id
        )
    )
    return json.loads(get_content(url))['http_mp3_128_url']


def soundcloud_download(url, info_only=False, **kwargs):
    url = 'https://api.soundcloud.com/resolve.json?url={}&client_id={}'.format(
        url, client_id
    )
    metadata = get_content(url)
    info = json.loads(metadata)
    title = info['title']
    real_url = info.get('download_url')
    if real_url is None:
        real_url = info.get('steram_url')
    if real_url is None:
        raise Exception('Cannot get media URI for {}'.format(url))
    real_url = soundcloud_i1_api(info['id'])
    mime, ext, size = url_info(real_url)
    print_info(site_info, title, mime, size)
    if not info_only:
        download_urls([real_url], title, ext, size, **kwargs)


download = soundcloud_download
download_playlist = playlist_not_supported(site_info)



# coding=utf-8

import re
import json

from lulu.common import (
    get_content,
    playlist_not_supported,
)
from lulu.extractor import SimpleExtractor


__all__ = ['douyin_download_by_url']
site_info = 'douyin'


class DouYin(SimpleExtractor):
    def extract(self, url, **kwargs):
        self.site_info = site_info
        page_content = get_content(url)
        match_rule = re.compile(r'var data = \[(.*?)\];')
        video_info = json.loads(match_rule.findall(page_content)[0])
        urls = [video_info['video']['play_addr']['url_list'][0]]
        if video_info['cha_list']:
            title = video_info['cha_list'][0]['cha_name']
        else:
            title = video_info['desc']

        return {
            'urls': urls,
            'title': title,
        }


download = douyin_download_by_url = DouYin()
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
from urllib import parse
from html import unescape

from lulu.common import (
    match1,
    get_head,
    url_info,
    print_info,
    get_content,
    download_urls,
    download_url_ffmpeg,
    playlist_not_supported,
)
from lulu.config import FAKE_HEADERS
from lulu.extractors.embed import embed_download


__all__ = ['universal_download']


def universal_download(
    url, output_dir='.', merge=True, info_only=False, **kwargs
):
    try:
        content_type = get_head(url, headers=FAKE_HEADERS)['Content-Type']
    except Exception:
        content_type = get_head(
            url, headers=FAKE_HEADERS, get_method='GET'
        )['Content-Type']
    if content_type.startswith('text/html'):
        try:
            embed_download(
                url, output_dir=output_dir, merge=merge, info_only=info_only,
                **kwargs
            )
        except Exception:
            pass
        else:
            return

    domains = url.split('/')[2].split('.')
    if len(domains) > 2:
        domains = domains[1:]
    site_info = '.'.join(domains)

    if content_type.startswith('text/html'):
        # extract an HTML page
        response = get_content(url)
        page = str(response)

        page_title = match1(page, r'<title>([^<]*)')
        if page_title:
            page_title = unescape(page_title)

        hls_urls = re.findall(
            r'(https?://[^;"\'\\]+' + '\.m3u8?' + r'[^;"\'\\]*)', page
        )
        if hls_urls:
            for hls_url in hls_urls:
                type_, ext, size = url_info(hls_url)
                print_info(site_info, page_title, type_, size)
                if not info_only:
                    download_url_ffmpeg(
                        url=hls_url, title=page_title, ext='mp4',
                        output_dir=output_dir
                    )
            return

        # most common media file extensions on the Internet
        media_exts = [
            '\.flv', '\.mp3', '\.mp4', '\.webm',
            '[-_]1\d\d\d\.jpe?g', '[-_][6-9]\d\d\.jpe?g',  # tumblr
            '[-_]1\d\d\dx[6-9]\d\d\.jpe?g',
            '[-_][6-9]\d\dx1\d\d\d\.jpe?g',
            '[-_][6-9]\d\dx[6-9]\d\d\.jpe?g',
            's1600/[\w%]+\.jpe?g',  # blogger
            'img[6-9]\d\d/[\w%]+\.jpe?g',  # oricon?
        ]

        urls = []
        for i in media_exts:
            urls += re.findall(
                r'(https?://[^;"\'\\]+' + i + r'[^;"\'\\]*)', page
            )

            p_urls = re.findall(
                r'(https?%3A%2F%2F[^;&]+' + i + r'[^;&]*)', page
            )
            urls += [parse.unquote(url) for url in p_urls]

            q_urls = re.findall(
                r'(https?:\\\\/\\\\/[^;"\']+' + i + r'[^;"\']*)', page
            )
            urls += [url.replace('\\\\/', '/') for url in q_urls]

        # a link href to an image is often an interesting one
        urls += re.findall(r'href="(https?://[^"]+\.jpe?g)"', page, re.I)
        urls += re.findall(r'href="(https?://[^"]+\.png)"', page, re.I)
        urls += re.findall(r'href="(https?://[^"]+\.gif)"', page, re.I)

        # MPEG-DASH MPD
        mpd_urls = re.findall(r'src="(https?://[^"]+\.mpd)"', page)
        for mpd_url in mpd_urls:
            cont = get_content(mpd_url)
            base_url = match1(cont, r'<BaseURL>(.*)</BaseURL>')
            urls += [match1(mpd_url, r'(.*/)[^/]*') + base_url]

        # have some candy!
        candies = []
        i = 1
        for url in set(urls):
            filename = parse.unquote(url.split('/')[-1])
            if 5 <= len(filename) <= 80:
                title = '.'.join(filename.split('.')[:-1])
            else:
                title = '{}'.format(i)
                i += 1

            candies.append({
                'url': url, 'title': title
            })

        for candy in candies:
            try:
                mime, ext, size = url_info(candy['url'], faker=True)
                if not size:
                    size = float('Int')
            except Exception:
                continue
            else:
                print_info(site_info, candy['title'], ext, size)
                if not info_only:
                    download_urls(
                        [candy['url']], candy['title'], ext, size,
                        output_dir=output_dir, merge=merge, faker=True
                    )
        return

    else:
        # direct download
        filename = parse.unquote(url.split('/')[-1])
        title = '.'.join(filename.split('.')[:-1])
        ext = filename.split('.')[-1]
        _, _, size = url_info(url, faker=True)
        print_info(site_info, title, ext, size)
        if not info_only:
            download_urls(
                [url], title, ext, size, output_dir=output_dir, merge=merge,
                faker=True
            )
        return


site_info = None
download = universal_download
download_playlist = playlist_not_supported('universal')



#!/usr/bin/env python

import json

from lulu.common import (
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['quanmin_download']
site_info = '全民直播 quanmin.tv'


def quanmin_download(url, info_only=False, **kwargs):
    roomid = url.split('/')[3].split('?')[0]

    json_request_url = (
        'https://m.quanmin.tv/json/rooms/{}/noinfo6.json'.format(roomid)
    )
    content = get_content(json_request_url)
    data = json.loads(content)

    title = data['title']

    if not data['play_status']:
        raise ValueError('The live stream is not online!')
    real_url = 'https://flv.quanmin.tv/live/{}.flv'.format(roomid)

    print_info(site_info, title, 'flv', float('inf'))
    if not info_only:
        download_urls([real_url], title, 'flv', None, **kwargs)


download = quanmin_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re

from lulu.common import (
    get_content,
    playlist_not_supported,
)
from lulu.extractor import SimpleExtractor


__all__ = ['kuaishou_download_by_url']
site_info = 'kuaishou.com'


class Kuaishou(SimpleExtractor):
    def extract(self, url, **kwargs):
        self.site_info = site_info
        page = get_content(url)
        title = re.search(
            r'<meta\s+property="og:title"\s+content="(.+?)"/>', page
        ).group(1)
        try:
            url = re.search(
                r'<meta\s+property="og:video:url"\s+content="(.+?)"/>', page
            ).group(1)
            file_format = url.split('.')[-1]
        except Exception:  # extract image
            url = re.search(
                r'<meta\s+property="og:image"\s+content="(.+?)"/>', page
            ).group(1)
            file_format = url.split('.')[-1]
        return {
            'title': title,
            'urls': [url],
            'file_format': file_format,
        }


download = kuaishou_download_by_url = Kuaishou()
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['vine_download']
site_info = 'Vine.co'


def vine_download(url, output_dir='.', merge=True, info_only=False, **kwargs):
    html = get_content(url)
    video_id = match1(url, r'vine.co/v/([^/]+)')
    title = match1(html, r'<title>([^<]*)</title>')
    stream = match1(
        html,
        r'<meta property="twitter:player:stream" content="([^"]*)">'
    )
    if not stream:  # https://vine.co/v/.../card
        stream = match1(html, r'"videoUrl":"([^"]+)"')
        if stream:
            stream = stream.replace('\\/', '/')
        else:
            posts_url = 'https://archive.vine.co/posts/{}.json'.format(
                video_id
            )
            json_data = json.loads(get_content(posts_url))
            stream = json_data['videoDashUrl']
            title = json_data['description']
            if title == '':
                title = '{}_{}'.format(
                    json_data['username'].replace(' ', '_'), video_id
                )

    mime, ext, size = url_info(stream)

    print_info(site_info, title, mime, size)
    if not info_only:
        download_urls([stream], title, ext, size, output_dir, merge=merge)


download = vine_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
)


__all__ = ['lizhi_download']
site_info = '荔枝 lizhi.fm'


# radio_id: e.g. 549759 from http://www.lizhi.fm/549759/
#
# Returns a list of tuples (audio_id, title, url) for each episode
# (audio) in the radio playlist. url is the direct link to the audio
# file.
def lizhi_extract_playlist_info(radio_id):
    # /api/radio_audios API parameters:
    #
    # - s: starting episode
    # - l: count (per page)
    # - band: radio_id
    #
    # We use l=65535 for poor man's pagination (that is, no pagination
    # at all -- hope all fits on a single page).
    #
    # TODO: Use /api/radio?band={radio_id} to get number of episodes
    # (au_cnt), then handle pagination properly.
    api_url = (
        'http://www.lizhi.fm/api/radio_audios?s=0&l=65535&band={}'.format(
            radio_id
        )
    )
    api_response = json.loads(get_content(api_url))
    return api_response


def lizhi_download_audio(audio_id, title, url, info_only=False, **kwargs):
    filetype, ext, size = url_info(url)
    print_info(site_info, title, filetype, size)
    if not info_only:
        download_urls([url], title, ext, size, **kwargs)


def lizhi_download_playlist(url, info_only=False, **kwargs):
    # Sample URL: http://www.lizhi.fm/549759/
    radio_id = match1(url, r'/(\d+)')
    if not radio_id:
        raise NotImplementedError('{} not supported'.format(url))
    for data in lizhi_extract_playlist_info(radio_id):
        lizhi_download_audio(
            data['id'], data['name'], data['fixedHighPlayUrl'],
            info_only=info_only, **kwargs
        )


def lizhi_download(url, output_dir='.', info_only=False, **kwargs):
    # Sample URL: http://www.lizhi.fm/549759/2508612053517249030
    m = re.search(r'/(?P<radio_id>\d+)/(?P<audio_id>\d+)', url)
    if not m:
        raise NotImplementedError('{} not supported'.format(url))
    radio_id = m.group('radio_id')
    audio_id = m.group('audio_id')
    # Look for the audio_id among the full list of episodes
    data = list(filter(
        lambda x: x['id'] == audio_id, lizhi_extract_playlist_info(radio_id)
    ))[0]
    lizhi_download_audio(
        audio_id, data['name'], data['fixedHighPlayUrl'], info_only=info_only,
        **kwargs
    )


download = lizhi_download
download_playlist = lizhi_download_playlist



#!/usr/bin/env python

import re
from html import unescape
from urllib import (
    parse,
    request,
)

from lulu.common import (
    match1,
    url_info,
    unicodize,
    print_info,
    get_content,
    get_location,
    download_urls,
    playlist_not_supported,
)
from lulu.config import YOUTUBE_CODECS


__all__ = ['google_download']
site_info = 'Google google.com'


fmt_level = dict(zip(
    [str(codec['itag']) for codec in YOUTUBE_CODECS],
    range(len(YOUTUBE_CODECS))
))


def google_download(url, info_only=False, **kwargs):
    # Percent-encoding Unicode URL
    url = parse.quote(url, safe=':/+%?=')

    service = url.split('/')[2].split('.')[0]

    if service == 'plus':  # Google Plus

        # attempt to extract images first
        # TBD: posts with > 4 images
        # TBD: album links
        html = get_content(parse.unquote(url))
        real_urls = []
        for src in re.findall(r'src="([^"]+)"[^>]*itemprop="image"', html):
            t = src.split('/')
            t[0], t[-2] = t[0] or 'https:', 's0-d'
            u = '/'.join(t)
            real_urls.append(u)
        if not real_urls:
            real_urls = [
                match1(html, r'<meta property="og:image" content="([^"]+)')
            ]
            real_urls = [re.sub(r'w\d+-h\d+-p', 's0', u) for u in real_urls]
        post_date = match1(html, r'"?(20\d\d[-/]?[01]\d[-/]?[0123]\d)"?')
        post_id = match1(html, r'/posts/([^"]+)')
        title = '{}_{}'.format(post_date, post_id)

        try:
            url = 'https://plus.google.com/{}'.format(
                match1(html, r'(photos/\d+/albums/\d+/\d+)\?authkey')
            )
            html = get_content(url)
            temp = re.findall(r'\[(\d+),\d+,\d+,"([^"]+)"\]', html)
            temp = sorted(temp, key=lambda x: fmt_level[x[0]])
            urls = [unicodize(i[1]) for i in temp if i[0] == temp[0][0]]
            assert urls
            real_urls = urls  # Look ma, there's really a video!

            post_url = match1(
                html, r'"(https://plus.google.com/[^/]+/posts/[^"]*)"'
            )
            post_author = match1(post_url, r'/\+([^/]+)/posts')
            if post_author:
                post_url = 'https://plus.google.com/+{}/posts/{}'.format(
                    parse.quote(post_author), match1(post_url, r'posts/(.+)')
                )
            post_html = get_content(post_url)
            title = match1(post_html, r'<title[^>]*>([^<\n]+)')
        except Exception:
            pass

        for i, real_url in enumerate(real_urls):
            title_i = '{}[{}]'.format(title, i) if len(real_urls) > 1 \
                else title
            _, ext, size = url_info(real_url)
            if ext is None:
                ext = 'mp4'

            print_info(site_info, title_i, ext, size)
            if not info_only:
                download_urls([real_url], title_i, ext, size, **kwargs)

    elif service in ['docs', 'drive']:  # Google Docs
        html = get_content(url)

        title = match1(html, r'"title":"([^"]*)"') or match1(
            html, r'<meta itemprop="name" content="([^"]*)"'
        )
        if len(title.split('.')) > 1:
            title = ".".join(title.split('.')[:-1])

        docid = match1(url, '/file/d/([^/]+)')

        request.install_opener(
            request.build_opener(request.HTTPCookieProcessor())
        )

        real_url = (
            'https://docs.google.com/uc?export=download&confirm=no_antivirus&'
            'id={}'.format(docid)
        )
        redirected_url = get_location(real_url)
        if real_url != redirected_url:
            # tiny file - get real url here
            _, ext, size = url_info(redirected_url)
            real_url = redirected_url
        else:
            # huge file - the real_url is a confirm page and real url is in it
            confirm_page = get_content(real_url)
            hrefs = re.findall(r'href="(.+?)"', confirm_page)
            for u in hrefs:
                if u.startswith('/uc?export=download'):
                    rel = unescape(u)
            confirm_url = 'https://docs.google.com' + rel
            real_url = get_location(confirm_url)
            _, ext, size = url_info(real_url)
            if size is None:
                size = 0

        print_info(site_info, title, ext, size)
        if not info_only:
            download_urls([real_url], title, ext, size, **kwargs)


download = google_download
download_playlist = playlist_not_supported(site_info)



# coding=utf-8

import re
import json

from lulu.common import (
    match1,
    get_content,
    playlist_not_supported,
)
from lulu.util import log
from lulu.extractor import VideoExtractor


site_info = '企鹅直播 live.qq.com'


class QiE(VideoExtractor):
    name = site_info

    # Last updated: 2015-11-24
    stream_types = [
        {'id': 'normal', 'container': 'flv', 'video_profile': '标清'},
        {'id': 'middle', 'container': 'flv', 'video_profile': '550'},
        {'id': 'middle2', 'container': 'flv', 'video_profile': '900'},
    ]

    id_dic = {i['video_profile']: (i['id']) for i in stream_types}

    api_endpoint = 'http://www.qie.tv/api/v1/room/{room_id}'
    game_ep = 'http://live.qq.com/game/game_details/get_game_details_info/'

    def get_room_id_from_url(self, match_id):
        meta = json.loads(get_content(self.game_ep + str(match_id)))
        if meta['error'] != 0:
            log.wtf('Error happens when accessing game_details api')
        rooms = meta['data']['anchor_data']
        for room in rooms:
            if room['is_use_room']:
                return room['room_id']
        log.wtf('No room available for match {}'.format(match_id))

    def get_vid_from_url(self, url):
        """Extracts video ID from live.qq.com.
        """
        hit = re.search(r'live.qq.com/(\d+)', url)
        if hit is not None:
            return hit.group(1)
        hit = re.search(r'live.qq.com/directory/match/(\d+)', url)
        if hit is not None:
            return self.get_room_id_from_url(hit.group(1))
        html = get_content(url)
        room_id = match1(html, r'room_id\":(\d+)')
        if room_id is None:
            log.wtf('Unknown page {}'.format(url))
        return room_id

    def download_playlist_by_url(self, url, **kwargs):
        pass

    def prepare(self, **kwargs):
        if self.url:
            self.vid = self.get_vid_from_url(self.url)
        self.referer = self.url
        content = get_content(self.api_endpoint.format(room_id=self.vid))
        content = json.loads(content)
        self.title = content['data']['room_name']
        rtmp_url = content['data']['rtmp_url']
        # stream_avalable = [i['name'] for i in content['data']['stream']]
        stream_available = {}
        stream_available['normal'] = '{}/{}'.format(
            rtmp_url, content['data']['rtmp_live']
        )
        if len(content['data']['rtmp_multi_bitrate']) > 0:
            for k, v in content['data']['rtmp_multi_bitrate'].items():
                stream_available[k] = rtmp_url + '/' + v

        for s in self.stream_types:
            if s['id'] in stream_available.keys():
                quality_id = s['id']
                url = stream_available[quality_id]
                self.streams[quality_id] = {
                    'container': 'flv',
                    'video_profile': s['video_profile'],
                    'size': 0,
                    'url': url
                }

    def extract(self, **kwargs):
        for i in self.streams:
            s = self.streams[i]
            s['src'] = [s['url']]
        if 'stream_id' in kwargs and kwargs['stream_id']:
            # Extract the stream
            stream_id = kwargs['stream_id']

            if stream_id not in self.streams:
                log.e('[Error] Invalid video format.')
                log.e(
                    'Run \'-i\' command with no specific video format to '
                    'view all available formats.'
                )
                exit(2)
        else:
            # Extract stream with the best quality
            stream_id = self.streams_sorted[0]['id']
            s['src'] = [s['url']]


site = QiE()
download = site.download_by_url
download_playlist = playlist_not_supported(site_info)



# coding=utf-8

from pathlib import Path

from lulu.common import (
    url_size,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.config import FAKE_HEADERS
from lulu.util.parser import get_parser


__all__ = ['pixivision_download']
site_info = 'pixivision pixivision.net'


def pixivision_download(url, output_dir='.', info_only=False, **kwargs):
    html = get_content(url)
    parser = get_parser(html)
    title = parser.h1.text.strip()
    output_dir = Path(output_dir) / title
    imgs = parser.find_all('img', class_='am__work__illust')
    print_info(site_info, title, 'jpg', 0)
    if not info_only:
        headers = FAKE_HEADERS.copy()
        headers.update({'Referer': url})
        for img in imgs:
            img = img['src']
            size = url_size(img, headers=headers)
            filename, ext = img.split('/')[-1].split('.')
            download_urls(
                [img], filename, ext, size, output_dir, refer=url, **kwargs
            )


download = pixivision_download
download_playlist = playlist_not_supported(site_info)



# coding=utf-8

import re
import json

from lulu.common import (
    url_size,
    get_content,
    playlist_not_supported,
)
from lulu.util import log
from lulu.extractors import VideoExtractor


__all__ = ['qingting_download_by_url']
site_info = '蜻蜓FM qingting.fm'


class Qingting(VideoExtractor):
    # every resource is described by its channel id and program id
    # so vid is tuple (chaanel_id, program_id)

    name = site_info
    stream_types = [
        {'id': '_default'}
    ]

    ep = 'http://i.qingting.fm/wapi/channels/{}/programs/{}'
    file_host = 'http://od.qingting.fm/{}'
    mobile_pt = r'channels\/(\d+)\/programs/(\d+)'

    def prepare(self, **kwargs):
        if self.vid is None:
            hit = re.search(self.__class__.mobile_pt, self.url)
            self.vid = (hit.group(1), hit.group(2))

        ep_url = self.__class__.ep.format(self.vid[0], self.vid[1])
        meta = json.loads(get_content(ep_url))

        if meta['code'] != 0:
            log.wtf(meta['message']['errormsg'])

        file_path = self.__class__.file_host.format(meta['data']['file_path'])
        self.title = meta['data']['name']
        duration = str(meta['data']['duration']) + 's'

        self.streams['_default'] = {
            'src': [file_path], 'video_profile': duration, 'container': 'm4a'
        }

    def extract(self, **kwargs):
        self.streams['_default']['size'] = url_size(
            self.streams['_default']['src'][0]
        )


def qingting_download_by_url(url, **kwargs):
    Qingting().download_by_url(url, **kwargs)


download = qingting_download_by_url
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re

from lulu.common import (
    match1,
    url_info,
    urls_size,
    unicodize,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['facebook_download']
site_info = 'facebook.com'


def facebook_download(
    url, output_dir='.', merge=True, info_only=False, **kwargs
):
    html = get_content(url)

    title = match1(html, r'<title id="pageTitle">(.+)</title>')

    if title is None:
        title = url

    sd_urls = list(set([
        unicodize(str.replace(i, '\\/', '/'))
        for i in re.findall(r'sd_src_no_ratelimit:"([^"]*)"', html)
    ]))
    hd_urls = list(set([
        unicodize(str.replace(i, '\\/', '/'))
        for i in re.findall(r'hd_src_no_ratelimit:"([^"]*)"', html)
    ]))
    urls = hd_urls if hd_urls else sd_urls

    _type, ext, size = url_info(urls[0], True)
    size = urls_size(urls)

    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls(urls, title, ext, size, output_dir, merge=False)


download = facebook_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

from pathlib import Path

from lulu.common import (
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.util.parser import get_parser


__all__ = ['bcy_download']
site_info = '半次元 bcy.net'


def bcy_download(url, output_dir='.', info_only=False, **kwargs):
    html = get_content(url)
    parser = get_parser(html)
    title = parser.h1.text.strip()
    output_dir = Path(output_dir) / title
    imgs = parser.find_all('img', class_='detail_std detail_clickable')
    print_info(site_info, title, 'jpg', 0)
    if not info_only:
        for img in imgs:
            # https://img9.bcyimg.com/drawer/15294/post/1799t/1f5a87801a0711e898b12b640777720f.jpg/w650  # noqa
            img = img['src'][:-5]
            filename, ext = img.split('/')[-1].split('.')
            download_urls(
                [img], filename, ext, 0, output_dir, **kwargs
            )


download = bcy_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

from lulu.common import (
    match1,
    get_content,
    any_download,
    playlist_not_supported,
)


__all__ = ['dilidili_download']
site_info = 'dilidili.com'


def dilidili_download(url, **kwargs):
    html = get_content(url)
    # player loaded via internal iframe
    # http://www.maoyun.tv/mdparse/index.php?id=http://v.youku.com/v_show/id_XMTYxNzk0NjUzMg==.html  # noqa
    iframe_url = match1(html, r'<iframe src=\"(.+?)\"')
    url = match1(
        iframe_url, r'https?://www.maoyun.tv/mdparse/index.php\?id=(.+)'
    )
    any_download(url, **kwargs)


download = dilidili_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import re
import json

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = [
    'ximalaya_download_playlist', 'ximalaya_download',
    'ximalaya_download_by_id'
]
site_info = '喜马拉雅FM ximalaya.com'


stream_types = [
    {'itag': '1', 'container': 'm4a', 'bitrate': 'default'},
    {'itag': '2', 'container': 'm4a', 'bitrate': '32'},
    {'itag': '3', 'container': 'm4a', 'bitrate': '64'},
]


def ximalaya_download_by_id(
    _id, title=None, info_only=False, stream_id=None, **kwargs
):
    BASE_URL = 'http://www.ximalaya.com/tracks/'
    json_url = '{}{}.json'.format(BASE_URL, _id)
    json_data = json.loads(get_content(json_url))
    if 'res' in json_data:
        if json_data['res'] is False:
            raise ValueError('Server reported id {} is invalid'.format(_id))
    if 'is_paid' in json_data and json_data['is_paid']:
        if 'is_free' in json_data and not json_data['is_free']:
            raise ValueError('{} is paid item'.format(_id))
    if (not title) and 'title' in json_data:
        title = json_data['title']
    # no size data in the json. should it be calculated?
    size = 0
    url = json_data['play_path_64']
    if stream_id:
        if stream_id == '1':
            url = json_data['play_path_32']
        elif stream_id == '0':
            url = json_data['play_path']
    ext = 'm4a'
    urls = [url]
    print('Site:        %s' % site_info)
    print('title:       %s' % title)
    if info_only:
        if stream_id:
            print_stream_info(stream_id)
        else:
            for item in range(0, len(stream_types)):
                print_stream_info(item)
    if not info_only:
        print('Type:        MPEG-4 audio m4a')
        print('Size:        N/A')
        download_urls(urls, title, ext, size, **kwargs)


def ximalaya_download(url, info_only=False, stream_id=None, **kwargs):
    if re.match(r'http://www\.ximalaya\.com/(\d+)/sound/(\d+)', url):
        _id = match1(url, r'http://www\.ximalaya\.com/\d+/sound/(\d+)')
    else:
        raise NotImplementedError(url)
    ximalaya_download_by_id(
        _id, info_only=info_only, stream_id=stream_id, **kwargs
    )


def ximalaya_download_page(
    playlist_url, info_only=False, stream_id=None, **kwargs
):
    if re.match(r'http://www\.ximalaya\.com/(\d+)/album/(\d+)', playlist_url):
        page_content = get_content(playlist_url)
        pattern = re.compile(r'<li sound_id="(\d+)"')
        ids = pattern.findall(page_content)
        for _id in ids:
            try:
                ximalaya_download_by_id(
                    _id, info_only=info_only, stream_id=stream_id, **kwargs
                )
            except(ValueError):
                print(
                    'Something wrong with {}, perhaps paid item?'.format(_id)
                )
    else:
        raise NotImplementedError(playlist_url)


def ximalaya_download_playlist(url, info_only=False, stream_id=None, **kwargs):
    match_result = re.match(
        r'http://www\.ximalaya\.com/(\d+)/album/(\d+)', url
    )
    if not match_result:
        raise NotImplementedError(url)
    pages = []
    page_content = get_content(url)
    if page_content.find('<div class="pagingBar_wrapper"') == -1:
        pages.append(url)
    else:
        base_url = 'http://www.ximalaya.com/{}/album/{}'.format(
            match_result.group(1), match_result.group(2)
        )
        html_str = '<a href=(\'|")\/{}\/album\/{}\?page='.format(
            match_result.group(1), match_result.group(2)
        )
        count = len(re.findall(html_str, page_content))
        for page_num in range(count):
            pages.append('{}?page={}'.format(base_url, str(page_num+1)))
            print(pages[-1])
    for page in pages:
        ximalaya_download_page(
            page, info_only=info_only, stream_id=stream_id, **kwargs
        )


def print_stream_info(stream_id):
    print('    - itag:        %s' % stream_id)
    print('      container:   %s' % 'm4a')
    print('      bitrate:     %s' % stream_types[int(stream_id)]['bitrate'])
    print('      size:        %s' % 'N/A')
    print('    # download-with: lulu --itag=%s [URL]' % stream_id)


download = ximalaya_download
download_playlist = ximalaya_download_playlist



#!/usr/bin/env python

import re
import json
from urllib import parse

from lulu.util import fs
from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls
)


__all__ = ['yinyuetai_download', 'yinyuetai_download_by_id']
site_info = 'YinYueTai.com'


def yinyuetai_download_by_id(
    vid, title=None, output_dir='.', merge=True, info_only=False
):
    video_info = json.loads(get_content(
        'http://www.yinyuetai.com/insite/get-video-info?json=true&'
        'videoId={}'.format(vid)
    ))
    url_models = video_info['videoInfo']['coreVideoInfo']['videoUrlModels']
    url_models = sorted(url_models, key=lambda i: i['qualityLevel'])
    url = url_models[-1]['videoUrl']
    _type = ext = match1(url, r'\.(flv|mp4)')
    _, _, size = url_info(url)

    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([url], title, ext, size, output_dir, merge=merge)


def yinyuetai_download(
    url, output_dir='.', merge=True, info_only=False, **kwargs
):
    _id = match1(url, r'http://\w+.yinyuetai.com/video/(\d+)') or match1(
        url, r'http://\w+.yinyuetai.com/video/h5/(\d+)'
    )
    if not _id:
        yinyuetai_download_playlist(
            url, output_dir=output_dir, merge=merge, info_only=info_only
        )
        return

    html = get_content(url)
    title = match1(
        html, r'<meta property="og:title"\s+content="([^"]+)"/>'
    ) or match1(html, r'<title>(.*)')
    assert title
    title = parse.unquote(title)
    title = fs.legitimize(title)
    yinyuetai_download_by_id(
        _id, title, output_dir, merge=merge, info_only=info_only
    )


def yinyuetai_download_playlist(
    url, output_dir='.', merge=True, info_only=False, **kwargs
):
    html = get_content(url)

    data_ids = re.findall(r'data-id="(\d+)"\s*data-index="\d+"', html)
    for data_id in data_ids:
        yinyuetai_download(
            'http://v.yinyuetai.com/video/{}'.format(data_id),
            output_dir=output_dir, merge=merge, info_only=info_only
        )


download = yinyuetai_download
download_playlist = yinyuetai_download_playlist



#!/usr/bin/env python

from urllib import (
    parse,
    request,
)

from lulu.common import (
    match1,
    url_info,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.config import FAKE_HEADERS


__all__ = ['nicovideo_download']
site_info = 'niconico nicovideo.jp'
fake_headers = FAKE_HEADERS.copy()


def nicovideo_login(user, password):
    data = 'current_form=login&mail={}&password={}&login_submit=Log+In'.format(
        user, password
    )
    response = request.urlopen(request.Request(
        'https://secure.nicovideo.jp/secure/login?site=niconico',
        headers=fake_headers, data=data.encode('utf-8')
    ))
    return response.headers


def nicovideo_download(url, info_only=False, **kwargs):
    import ssl
    ssl_context = request.HTTPSHandler(
        context=ssl.SSLContext(ssl.PROTOCOL_TLSv1)
    )
    cookie_handler = request.HTTPCookieProcessor()
    opener = request.build_opener(ssl_context, cookie_handler)
    request.install_opener(opener)

    import netrc
    import getpass
    try:
        info = netrc.netrc().authenticators('nicovideo')
    except Exception:
        info = None
    if info is None:
        user = input('User:     ')
        password = getpass.getpass('Password: ')
    else:
        user, password = info[0], info[2]
    print('Logging in...')
    nicovideo_login(user, password)

    html = get_content(url)  # necessary!
    title = match1(html, r'<title>(.+?)</title>')

    vid = url.split('/')[-1].split('?')[0]
    api_html = get_content('http://flapi.nicovideo.jp/api/getflv?v={}'.format(
        vid
    ))
    real_url = parse.unquote(match1(api_html, r'url=([^&]+)&'))

    _type, ext, size = url_info(real_url)
    print_info(site_info, title, _type, size)
    if not info_only:
        download_urls([real_url], title, ext, size, **kwargs)


download = nicovideo_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import ssl
from urllib import request

from lulu.common import (
    match1,
    url_info,
    get_content,
)
from lulu.extractor import VideoExtractor


class Infoq(VideoExtractor):
    name = 'InfoQ infoq.com'

    stream_types = [
        {'id': 'video'},
        {'id': 'audio'},
        {'id': 'slides'}
    ]

    def prepare(self, **kwargs):
        content = get_content(self.url)
        self.title = match1(content, r'<title>([^<]+)</title>')
        s = match1(content, r'P\.s\s*=\s*\'([^\']+)\'')
        scp = match1(content, r'InfoQConstants\.scp\s*=\s*\'([^\']+)\'')
        scs = match1(content, r'InfoQConstants\.scs\s*=\s*\'([^\']+)\'')
        sck = match1(content, r'InfoQConstants\.sck\s*=\s*\'([^\']+)\'')

        mp3 = match1(content, r'name="filename"\s*value="([^"]+\.mp3)"')
        if mp3:
            mp3 = 'http://res.infoq.com/downloads/mp3downloads/{}'.format(mp3)

        pdf = match1(content, r'name="filename"\s*value="([^"]+\.pdf)"')
        if pdf:
            pdf = 'http://res.infoq.com/downloads/pdfdownloads/{}'.format(pdf)

        # cookie handler
        ssl_context = request.HTTPSHandler(
            context=ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        )
        cookie_handler = request.HTTPCookieProcessor()
        opener = request.build_opener(ssl_context, cookie_handler)
        opener.addheaders = [
            ('Referer', self.url),
            (
                'Cookie',
                'CloudFront-Policy={};CloudFront-Signature={};'
                'CloudFront-Key-Pair-Id={}'.format(scp, scs, sck)
            )
        ]
        request.install_opener(opener)
        if s:
            self.streams['video'] = {'url': s}
        if mp3:
            self.streams['audio'] = {'url': mp3}
        if pdf:
            self.streams['slides'] = {'url': pdf}

    def extract(self, **kwargs):
        for i in self.streams:
            s = self.streams[i]
            _, s['container'], s['size'] = url_info(s['url'])
            s['src'] = [s['url']]


site = Infoq()
download = site.download_by_url
download_playlist = site.download_by_url



#!/usr/bin/env python

import re
import json
import time

from lulu.util import log
from lulu.common import (
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)


__all__ = ['panda_download']
site_info = '熊猫直播 panda.tv'


def panda_download(url, info_only=False, **kwargs):
    roomid = re.search('/(\d+)', url)
    if roomid is None:
        log.wtf('Cannot found room id for this url')
    roomid = roomid.group(1)
    json_request_url = (
        'http://www.panda.tv/api_room_v2?roomid={}&__plat=pc_web&_={}'.format(
            roomid, int(time.time())
        )
    )
    content = get_content(json_request_url)
    api_json = json.loads(content)

    errno = api_json['errno']
    errmsg = api_json['errmsg']
    if errno:
        raise ValueError('Errno : {}, Errmsg : {}'.format(errno, errmsg))
    data = api_json['data']
    title = data['roominfo']['name']
    room_key = data['videoinfo']['room_key']
    plflag = data['videoinfo']['plflag'].split('_')
    status = data['videoinfo']['status']
    if status is not '2':
        raise ValueError('The live stream is not online! (status:{})'.format(
            status)
        )

    data2 = json.loads(data['videoinfo']['plflag_list'])
    rid = data2['auth']['rid']
    sign = data2['auth']['sign']
    ts = data2['auth']['time']
    real_url = (
        'http://pl{}.live.panda.tv/live_panda/{}.flv?sign={}&ts={}&'
        'rid={}'.format(plflag[1], room_key, sign, ts, rid)
    )
    print_info(site_info, title, 'flv', float('inf'))
    if not info_only:
        download_urls([real_url], title, 'flv', None, **kwargs)


download = panda_download
download_playlist = playlist_not_supported(site_info)



#!/usr/bin/env python

import json

from lulu.common import (
    match1,
    get_content,
)
from lulu.util import log
from lulu.extractors.bokecc import bokecc_download_by_id


__all__ = [
    'wanmen_download', 'wanmen_download_by_course',
    'wanmen_download_by_course_topic',
    'wanmen_download_by_course_topic_part'
]
site_info = '万门 wanmen.org'


# Helper functions
def _wanmen_get_json_api_content_by_courseID(courseID):
    """int->JSON

    Return a parsed JSON tree of WanMen's API.
    """

    return json.loads(get_content(
        'http://api.wanmen.org/course/getCourseNested/{courseID}'.format(
            courseID=courseID)
        ))


def _wanmen_get_title_by_json_topic_part(json_content, tIndex, pIndex):
    """JSON, int, int, int->str

    Get a proper title with courseid+topicID+partID.
    """

    return '_'.join([
        json_content[0]['name'],
        json_content[0]['Topics'][tIndex]['name'],
        json_content[0]['Topics'][tIndex]['Parts'][pIndex]['name']
    ])


def _wanmen_get_boke_id_by_json_topic_part(json_content, tIndex, pIndex):
    """JSON, int, int, int->str

    Get one BokeCC video ID with courseid+topicID+partID."""

    return json_content[0]['Topics'][tIndex]['Parts'][pIndex]['ccVideoLink']


# Parsers
def wanmen_download_by_course(json_api_content, info_only=False, **kwargs):
    """int->None

    Download a WHOLE course.
    Reuse the API call to save time.
    """

    for tIndex in range(len(json_api_content[0]['Topics'])):
        for pIndex in range(
            len(json_api_content[0]['Topics'][tIndex]['Parts'])
        ):
            wanmen_download_by_course_topic_part(
                json_api_content, tIndex, pIndex, info_only=info_only,
                **kwargs
            )


def wanmen_download_by_course_topic(
    json_api_content, tIndex, info_only=False, **kwargs
):
    """int, int->None

    Download a TOPIC of a course.
    Reuse the API call to save time.
    """

    for pIndex in range(len(json_api_content[0]['Topics'][tIndex]['Parts'])):
        wanmen_download_by_course_topic_part(
            json_api_content, tIndex, pIndex, info_only=info_only, **kwargs
        )


def wanmen_download_by_course_topic_part(
    json_api_content, tIndex, pIndex, info_only=False, **kwargs
):
    """int, int, int->None

    Download ONE PART of the course.
    """

    html = json_api_content

    title = _wanmen_get_title_by_json_topic_part(
        html, tIndex, pIndex
    )
    bokeccID = _wanmen_get_boke_id_by_json_topic_part(
        html, tIndex, pIndex
    )
    bokecc_download_by_id(
        vid=bokeccID, title=title, info_only=info_only, **kwargs
    )


# Main entrance
def wanmen_download(url, info_only=False, **kwargs):

    if 'wanmen.org' not in url:
        log.wtf(
            'You are at the wrong place dude. This is for WanMen University!'
        )

    courseID = int(match1(url, r'course\/(\d+)'))
    assert courseID > 0  # without courseID we cannot do anything

    tIndex = int(match1(url, r'tIndex=(\d+)'))

    pIndex = int(match1(url, r'pIndex=(\d+)'))

    json_api_content = _wanmen_get_json_api_content_by_courseID(courseID)

    if pIndex:  # only download ONE single part
        assert tIndex >= 0
        wanmen_download_by_course_topic_part(
            json_api_content, tIndex, pIndex, info_only=info_only, **kwargs
        )
    elif tIndex:  # download a topic
        wanmen_download_by_course_topic(
            json_api_content, tIndex, info_only=info_only, **kwargs
        )
    else:  # download the whole course
        wanmen_download_by_course(
            json_api_content, info_only=info_only, **kwargs
        )


download = wanmen_download
download_playlist = wanmen_download_by_course



#!/usr/bin/env python


from lulu.common import (
    get_content,
    playlist_not_supported,
)
from lulu.util.parser import get_parser
from lulu.extractors.youtube import YouTube


__all__ = ['khan_download']
site_info = 'Khan Academy khanacademy.org'


def khan_download(url, **kwargs):
    html = get_content(url)
    parser = get_parser(html)
    youtube_url = parser.find('meta', property='og:video')['content']
    YouTube().download_by_url(youtube_url, **kwargs)


download = khan_download
download_playlist = playlist_not_supported(site_info)



# coding=utf-8

SITES = {
    '163': 'netease',
    '56': 'w56',
    'acfun': 'acfun',
    'archive': 'archive',
    'baidu': 'baidu',
    'bandcamp': 'bandcamp',
    'baomihua': 'baomihua',
    'bcy': 'bcy',
    'bigthink': 'bigthink',
    'bilibili': 'bilibili',
    'cctv': 'cntv',
    'cntv': 'cntv',
    'cbs': 'cbs',
    'coub': 'coub',
    'dailymotion': 'dailymotion',
    'dilidili': 'dilidili',
    'douban': 'douban',
    'douyin': 'douyin',
    'douyu': 'douyutv',
    'ehow': 'ehow',
    'facebook': 'facebook',
    'fantasy': 'fantasy',
    'fc2': 'fc2video',
    'flickr': 'flickr',
    'freesound': 'freesound',
    'fun': 'funshion',
    'google': 'google',
    'giphy': 'giphy',
    'heavy-music': 'heavymusic',
    'huaban': 'huaban',
    'huomao': 'huomaotv',
    'iask': 'sina',
    'icourses': 'icourses',
    'ifeng': 'ifeng',
    'imgur': 'imgur',
    'in': 'alive',
    'infoq': 'infoq',
    'instagram': 'instagram',
    'iqilu': 'iqilu',
    'iqiyi': 'iqiyi',
    'ixigua': 'ixigua',
    'isuntv': 'suntv',
    'joy': 'joy',
    'kankanews': 'bilibili',
    'khanacademy': 'khan',
    'ku6': 'ku6',
    'kuaishou': 'kuaishou',
    'kugou': 'kugou',
    'kuwo': 'kuwo',
    'le': 'le',
    'letv': 'le',
    'lizhi': 'lizhi',
    'longzhu': 'longzhu',
    'magisto': 'magisto',
    'metacafe': 'metacafe',
    'mgtv': 'mgtv',
    'mtv81': 'mtv81',
    'musicplayon': 'musicplayon',
    'naver': 'naver',
    '7gogo': 'nanagogo',
    'nicovideo': 'nicovideo',
    'panda': 'panda',
    'pinterest': 'pinterest',
    'pixivision': 'pixivision',
    'pixnet': 'pixnet',
    'pptv': 'pptv',
    'qingting': 'qingting',
    'qq': 'qq',
    'quanmin': 'quanmin',
    'showroom-live': 'showroom',
    'sina': 'sina',
    'smgbb': 'bilibili',
    'sohu': 'sohu',
    'soundcloud': 'soundcloud',
    'ted': 'ted',
    'theplatform': 'theplatform',
    'tucao': 'tucao',
    'tudou': 'tudou',
    'tumblr': 'tumblr',
    'twimg': 'twitter',
    'twitter': 'twitter',
    'ucas': 'ucas',
    'videomega': 'videomega',
    'vidto': 'vidto',
    'vimeo': 'vimeo',
    'wanmen': 'wanmen',
    'weibo': 'weibo',
    'veoh': 'veoh',
    'vine': 'vine',
    'vk': 'vk',
    'xiami': 'xiami',
    'xiaokaxiu': 'yixia',
    'xiaojiadianvideo': 'fc2video',
    'ximalaya': 'ximalaya',
    'yinyuetai': 'yinyuetai',
    'miaopai': 'yixia',
    'yizhibo': 'yizhibo',
    'youku': 'youku',
    'iwara': 'iwara',
    'youtu': 'youtube',
    'youtube': 'youtube',
    'zhanqi': 'zhanqi',
    '365yg': 'toutiao',
}

FAKE_HEADERS = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',  # noqa
    'Accept-Charset': 'UTF-8,*;q=0.5',
    'Accept-Encoding': 'gzip,deflate,sdch',
    'Accept-Language': 'en-US,en;q=0.8',
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:51.0) Gecko/20100101 Firefox/51.0',  # noqa
}

FAKE_HEADERS_MOBILE = {
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',  # noqa
    'Accept-Charset': 'UTF-8,*;q=0.5',
    'Accept-Encoding': 'gzip,deflate,sdch',
    'Accept-Language': 'en-US,en;q=0.8',
    'User-Agent': 'Mozilla/5.0 (Linux; Android 4.4.2; Nexus 4 Build/KOT49H) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.114 Mobile Safari/537.36',  # noqa
}

# netease
NETEASE_MP3_URL = (
    'http://music.163.com/weapi/song/enhance/player/url?csrf_token='
)
NETEASE_MUSIC_PUBKEY = '010001'
NETEASE_MUSIC_SECKEY = 16 * 'F'
NETEASE_MUSIC_COMMENT_MODULE = '00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7'  # noqa

# YouTube
# YouTube media encoding options, in descending quality order.
# http://en.wikipedia.org/wiki/YouTube#Quality_and_codecs.
# Retrieved July 17, 2014.
YOUTUBE_CODECS = [
    {
        'itag': '38', 'container': 'MP4', 'video_resolution': '3072p',
        'video_encoding': 'H.264', 'video_profile': 'High',
        'video_bitrate': '3.5-5', 'audio_encoding': 'AAC',
        'audio_bitrate': '192',
    },
    # {'itag': '85', 'container': 'MP4', 'video_resolution': '1080p', 'video_encoding': 'H.264', 'video_profile': '3D', 'video_bitrate': '3-4', 'audio_encoding': 'AAC', 'audio_bitrate': '192'},  # noqa
    {
        'itag': '46', 'container': 'WebM', 'video_resolution': '1080p',
        'video_encoding': 'VP8', 'video_profile': '', 'video_bitrate': '',
        'audio_encoding': 'Vorbis', 'audio_bitrate': '192',
    },
    {
        'itag': '37', 'container': 'MP4', 'video_resolution': '1080p',
        'video_encoding': 'H.264', 'video_profile': 'High',
        'video_bitrate': '3-4.3', 'audio_encoding': 'AAC',
        'audio_bitrate': '192',
    },
    # {'itag': '102', 'container': 'WebM', 'video_resolution': '720p', 'video_encoding': 'VP8', 'video_profile': '3D', 'video_bitrate': '', 'audio_encoding': 'Vorbis', 'audio_bitrate': '192'},  # noqa
    {
        'itag': '45', 'container': 'WebM', 'video_resolution': '720p',
        'video_encoding': 'VP8', 'video_profile': '', 'video_bitrate': '2',
        'audio_encoding': 'Vorbis', 'audio_bitrate': '192',
    },
    # {'itag': '84', 'container': 'MP4', 'video_resolution': '720p', 'video_encoding': 'H.264', 'video_profile': '3D', 'video_bitrate': '2-3', 'audio_encoding': 'AAC', 'audio_bitrate': '192'},  # noqa
    {
        'itag': '22', 'container': 'MP4', 'video_resolution': '720p',
        'video_encoding': 'H.264', 'video_profile': 'High',
        'video_bitrate': '2-3', 'audio_encoding': 'AAC',
        'audio_bitrate': '192',
    },
    {
        'itag': '120', 'container': 'FLV', 'video_resolution': '720p',
        'video_encoding': 'H.264', 'video_profile': 'Main@L3.1',
        'video_bitrate': '2', 'audio_encoding': 'AAC',
        'audio_bitrate': '128',
    },  # Live streaming only
    {
        'itag': '44', 'container': 'WebM', 'video_resolution': '480p',
        'video_encoding': 'VP8', 'video_profile': '', 'video_bitrate': '1',
        'audio_encoding': 'Vorbis', 'audio_bitrate': '128',
    },
    {
        'itag': '35', 'container': 'FLV', 'video_resolution': '480p',
        'video_encoding': 'H.264', 'video_profile': 'Main',
        'video_bitrate': '0.8-1', 'audio_encoding': 'AAC',
        'audio_bitrate': '128',
    },
    # {'itag': '101', 'container': 'WebM', 'video_resolution': '360p', 'video_encoding': 'VP8', 'video_profile': '3D', 'video_bitrate': '', 'audio_encoding': 'Vorbis', 'audio_bitrate': '192'},  # noqa
    # {'itag': '100', 'container': 'WebM', 'video_resolution': '360p', 'video_encoding': 'VP8', 'video_profile': '3D', 'video_bitrate': '', 'audio_encoding': 'Vorbis', 'audio_bitrate': '128'},  # noqa
    {
        'itag': '43', 'container': 'WebM', 'video_resolution': '360p',
        'video_encoding': 'VP8', 'video_profile': '',
        'video_bitrate': '0.5', 'audio_encoding': 'Vorbis',
        'audio_bitrate': '128',
    },
    {
        'itag': '34', 'container': 'FLV', 'video_resolution': '360p',
        'video_encoding': 'H.264', 'video_profile': 'Main',
        'video_bitrate': '0.5', 'audio_encoding': 'AAC',
        'audio_bitrate': '128',
    },
    # {'itag': '82', 'container': 'MP4', 'video_resolution': '360p', 'video_encoding': 'H.264', 'video_profile': '3D', 'video_bitrate': '0.5', 'audio_encoding': 'AAC', 'audio_bitrate': '96'},  # noqa
    {
        'itag': '18', 'container': 'MP4', 'video_resolution': '270p/360p',
        'video_encoding': 'H.264', 'video_profile': 'Baseline',
        'video_bitrate': '0.5', 'audio_encoding': 'AAC',
        'audio_bitrate': '96',
    },
    {
        'itag': '6', 'container': 'FLV', 'video_resolution': '270p',
        'video_encoding': 'Sorenson H.263', 'video_profile': '',
        'video_bitrate': '0.8', 'audio_encoding': 'MP3',
        'audio_bitrate': '64',
    },
    # {'itag': '83', 'container': 'MP4', 'video_resolution': '240p', 'video_encoding': 'H.264', 'video_profile': '3D', 'video_bitrate': '0.5', 'audio_encoding': 'AAC', 'audio_bitrate': '96'},  # noqa
    {
        'itag': '13', 'container': '3GP', 'video_resolution': '',
        'video_encoding': 'MPEG-4 Visual', 'video_profile': '',
        'video_bitrate': '0.5', 'audio_encoding': 'AAC',
        'audio_bitrate': '',
    },
    {
        'itag': '5', 'container': 'FLV', 'video_resolution': '240p',
        'video_encoding': 'Sorenson H.263', 'video_profile': '',
        'video_bitrate': '0.25', 'audio_encoding': 'MP3',
        'audio_bitrate': '64',
    },
    {
        'itag': '36', 'container': '3GP', 'video_resolution': '240p',
        'video_encoding': 'MPEG-4 Visual', 'video_profile': 'Simple',
        'video_bitrate': '0.175', 'audio_encoding': 'AAC',
        'audio_bitrate': '36',
    },
    {
        'itag': '17', 'container': '3GP', 'video_resolution': '144p',
        'video_encoding': 'MPEG-4 Visual', 'video_profile': 'Simple',
        'video_bitrate': '0.05', 'audio_encoding': 'AAC',
        'audio_bitrate': '24',
    },
]



#!/usr/bin/env python

import re
import json
from urllib import parse
from html import unescape

from lulu.common import (
    match1,
    get_head,
    urls_size,
    print_info,
    get_content,
    download_urls,
    playlist_not_supported,
)
from lulu.extractors.vine import vine_download


__all__ = ['twitter_download']
site_info = 'Twitter twitter.com'


def extract_m3u(source):
    r1 = get_content(source)
    s1 = re.findall(r'(/ext_tw_video/.*)', r1)
    s1 += re.findall(r'(/amplify_video/.*)', r1)
    r2 = get_content('https://video.twimg.com{}'.format(s1[-1]))
    s2 = re.findall(r'(/ext_tw_video/.*)', r2)
    s2 += re.findall(r'(/amplify_video/.*)', r2)
    return ['https://video.twimg.com{}'.format(i) for i in s2]


def twitter_download(url, info_only=False, **kwargs):
    html = get_content(url)
    screen_name = match1(html, r'data-screen-name="([^"]*)"') or \
        match1(html, r'<meta name="twitter:title" content="([^"]*)"')
    item_id = match1(html, r'data-item-id="([^"]*)"') or \
        match1(html, r'<meta name="twitter:site:id" content="([^"]*)"')
    page_title = '{} [{}]'.format(screen_name, item_id)

    try:  # extract images
        urls = re.findall(
            r'property="og:image"\s*content="([^"]+:large)"', html
        )
        assert urls
        images = []
        for url in urls:
            url = ':'.join(url.split(':')[:-1]) + ':orig'
            filename = parse.unquote(url.split('/')[-1])
            title = '.'.join(filename.split('.')[:-1])
            ext = url.split(':')[-2].split('.')[-1]
            size = int(get_head(url)['Content-Length'])
            images.append({
                'title': title,
                'url': url,
                'ext': ext,
                'size': size
            })
        size = sum([image['size'] for image in images])
        print_info(site_info, page_title, images[0]['ext'], size)

        if not info_only:
            for image in images:
                title = image['title']
                ext = image['ext']
                size = image['size']
                url = image['url']
                print_info(site_info, title, ext, size)
                download_urls([url], title, ext, size, **kwargs)

    except Exception:  # extract video
        # always use i/cards or videos url
        if not re.match(r'https?://twitter.com/i/', url):
            url = match1(
                html, r'<meta\s*property="og:video:url"\s*content="([^"]+)"'
            )
            if not url:
                url = 'https://twitter.com/i/videos/{}'.format(item_id)
            html = get_content(url)

        data_config = match1(html, r'data-config="([^"]*)"') or \
            match1(html, r'data-player-config="([^"]*)"')
        i = json.loads(unescape(data_config))
        if 'video_url' in i:
            source = i['video_url']
            item_id = i['tweet_id']
            page_title = "{} [{}]".format(screen_name, item_id)
        elif 'playlist' in i:
            source = i['playlist'][0]['source']
            if not item_id:
                page_title = i['playlist'][0]['contentId']
        elif 'vmap_url' in i:
            vmap_url = i['vmap_url']
            vmap = get_content(vmap_url)
            source = match1(vmap, r'<MediaFile>\s*<!\[CDATA\[(.*)\]\]>')
            item_id = i['tweet_id']
            page_title = '{} [{}]'.format(screen_name, item_id)
        elif 'scribe_playlist_url' in i:
            scribe_playlist_url = i['scribe_playlist_url']
            return vine_download(
                scribe_playlist_url, info_only=info_only, **kwargs
            )

        try:
            urls = extract_m3u(source)
        except Exception:
            urls = [source]
        size = urls_size(urls)
        mime, ext = 'video/mp4', 'mp4'

        print_info(site_info, page_title, mime, size)
        if not info_only:
            download_urls(urls, page_title, ext, size, **kwargs)


download = twitter_download
download_playlist = playlist_not_supported(site_info)
